# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP 'Electricity' worksheet."""

from __future__ import annotations
from typing import List

from PHX.to_PHPP import xl_app, xl_data
from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP.phpp_model import electricity_item

class Electricity:
    """IO Controller for PHPP "Electricity" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Electricity):
        self.xl = _xl
        self.shape = shape

    def _turn_off_all_equipment(self):
        """Sets all the 'used' values to 0 to reset the sheet before writing new equipment."""
        for item in self.shape.input_rows:
            # Some items cannot be turned off....
            excluded = ['clothes_drying', 'cooking', 'consumer_elec', 'lighting', 'small_appliances']
            if item[0] in excluded:
                continue 

            self.xl.write_xl_item(
                    xl_data.XlItem(
                        self.shape.name,
                        f'{self.shape.input_columns.used}{item[1].data}',
                        0
                    )
                )

    def write_equipment(self, _equipment_inputs: List[electricity_item.ElectricityItem]) -> None:
        """Write a list of equipment-input objects to the Worksheet."""
        self._turn_off_all_equipment()
        
        for equip_input in _equipment_inputs:
            for item in equip_input.create_xl_items(self.shape):
                self.xl.write_xl_item(item)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Passive House Certification Classes"""

from __future__ import annotations
from typing import ClassVar
from dataclasses import dataclass, field
from enum import Enum

from PHX.model import ground
from PHX.model.enums import phi_certification_phpp_9, phius_certification


@dataclass
class PhxSetpoints:
    winter: float = 20  # deg. C
    summer: float = 25  # deg. C


@dataclass
class PhxPhBuildingData:
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    
    num_of_units: int = 1
    num_of_floors: int = 1
    occupancy_setting_method: int = 2  # Design

    airtightness_q50: float = 1.0  # m3/hr-m2-envelope
    airtightness_n50: float = 1.0  # ach
    wind_coefficient_e: float = 0.07
    wind_coefficient_f: float = 15
    
    setpoints: PhxSetpoints = field(default_factory=PhxSetpoints)
    
    foundations: list[ground.PhxFoundation] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def add_foundation(self, _input: ground.PhxFoundation) -> None:
        self.foundations.append(_input)


# -----------------------------------------------------------------------------
@dataclass
class PhxPhiusCertificationCriteria:
    ph_certificate_criteria: int = 3
    ph_selection_target_data: int = 2
    
    phius_annual_heating_demand: float = 15.0
    phius_annual_cooling_demand: float = 15.0
    phius_peak_heating_load: float = 10.0
    phius_peak_cooling_load: float = 10.0

@dataclass
class PhxPhiusCertificationSettings:
    phius_building_category_type = phius_certification.PhiusCertificationBuildingCategoryType.RESIDENTIAL_BUILDING
    phius_building_use_type = phius_certification.PhiusCertificationBuildingUseType.RESIDENTIAL
    phius_building_status = phius_certification.PhiusCertificationBuildingStatus.IN_PLANNING
    phius_building_type = phius_certification.PhiusCertificationBuildingType.NEW_CONSTRUCTION

@dataclass
class PhxPhiusCertification:
    phius_certification_criteria: PhxPhiusCertificationCriteria = field(
        default_factory=PhxPhiusCertificationCriteria)
    phius_certification_settings: PhxPhiusCertificationSettings = field(
        default_factory=PhxPhiusCertificationSettings)
    
    # TODO: Refactor this out to someplace more general than inside Phius....
    ph_building_data: PhxPhBuildingData = field(
        default_factory=PhxPhBuildingData)


# -----------------------------------------------------------------------------
@dataclass
class PhxPhiCertificationSettings:
    phi_building_category_type: Enum = phi_certification_phpp_9.PhiCertBuildingCategoryType.RESIDENTIAL_BUILDING
    phi_building_use_type: Enum = phi_certification_phpp_9.PhiCertBuildingUseType.DWELLING
    phi_building_ihg_type: Enum = phi_certification_phpp_9.PhiCertIHGType.STANDARD
    phi_building_occupancy_type: Enum = phi_certification_phpp_9.PhiCertOccupancyType.STANDARD

    phi_certification_type: Enum = phi_certification_phpp_9.PhiCertType.PASSIVE_HOUSE
    phi_certification_class: Enum = phi_certification_phpp_9.PhiCertClass.CLASSIC
    phi_pe_type: Enum = phi_certification_phpp_9.PhiCertificationPEType.PER
    phi_enerphit_type: Enum = phi_certification_phpp_9.PhiCertEnerPHitType.BY_DEMAND
    phi_retrofit_type: Enum = phi_certification_phpp_9.PhiCertRetrofitType.NEW_BUILDING

@dataclass
class PhxPhiCertification:
    phi_certification_settings: PhxPhiCertificationSettings = field(
        default_factory=PhxPhiCertificationSettings)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Site (Location and Climate) Dataclasses"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Union, Optional

from PHX.model.enums.phx_site import SiteSelection, SiteClimateSelection, SiteEnergyFactorSelection


@dataclass
class PhxGround:
    ground_thermal_conductivity: float = 2
    ground_heat_capacity: float = 1000
    ground_density: float = 2000
    depth_groundwater: float = 3
    flow_rate_groundwater: float = 0.05


@dataclass
class PhxPEFactor:
    """Conversion Factors for Site-Energy->Primary-Energy"""
    value: float = 0.0
    unit: str = ''
    fuel_name: str = ''


@dataclass
class PhxCO2Factor:
    """Conversion Factors for Site->CO2"""
    value: float = 0.0
    unit: str = ''
    fuel_name: str = ''


PhxEnergyFactorAlias = Union[PhxPEFactor, PhxCO2Factor]


@dataclass
class PhxSiteEnergyFactors:
    selection_pe_co2_factor: SiteEnergyFactorSelection = SiteEnergyFactorSelection.USER_DEFINED
    pe_factors: dict[str, PhxEnergyFactorAlias] = field(default_factory=dict)
    co2_factors: dict[str, PhxEnergyFactorAlias] = field(default_factory=dict)

    def __post_init__(self):
        self.pe_factors = {
            "OIL": PhxPEFactor(1.1, "kWh/kWh", "OIL"),
            "NATURAL_GAS": PhxPEFactor(1.1, "kWh/kWh", "NATURAL_GAS"),
            "LPG": PhxPEFactor(1.1, "kWh/kWh", "LPG"),
            "HARD_COAL": PhxPEFactor(1.1, "kWh/kWh", "HARD_COAL"),
            "WOOD": PhxPEFactor(0.2, "kWh/kWh", "WOOD"),
            "ELECTRICITY_MIX": PhxPEFactor(1.8, "kWh/kWh", "ELECTRICITY_MIX"),
            "ELECTRICITY_PV": PhxPEFactor(1.7, "kWh/kWh", "ELECTRICITY_PV"),
            "HARD_COAL_CGS_70_CHP": PhxPEFactor(0.8, "kWh/kWh", "HARD_COAL_CGS_70_CHP"),
            "HARD_COAL_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "HARD_COAL_CGS_35_CHP"),
            "HARD_COAL_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "HARD_COAL_CGS_0_CHP"),
            "GAS_CGS_70_CHP": PhxPEFactor(0.7, "kWh/kWh", "GAS_CGS_70_CHP"),
            "GAS_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "GAS_CGS_35_CHP"),
            "GAS_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "GAS_CGS_0_CHP"),
            "OIL_CGS_70_CHP": PhxPEFactor(0.8, "kWh/kWh", "OIL_CGS_70_CHP"),
            "OIL_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "OIL_CGS_35_CHP"),
            "OIL_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "OIL_CGS_0_CHP"),
        }
        self.co2_factors: dict[str, PhxEnergyFactorAlias] = {
            "OIL": PhxCO2Factor(309.9966, "g/kWh", "OIL"),
            "NATURAL_GAS": PhxCO2Factor(250.0171, "g/kWh", "NATURAL_GAS"),
            "LPG": PhxCO2Factor(270.0102, "g/kWh", "LPG"),
            "HARD_COAL": PhxCO2Factor(439.9864, "g/kWh", "HARD_COAL"),
            "WOOD": PhxCO2Factor(53.4289, "g/kWh", "WOOD"),
            "ELECTRICITY_MIX": PhxCO2Factor(680.0068, "g/kWh", "ELECTRICITY_MIX"),
            "ELECTRICITY_PV": PhxCO2Factor(250.0171, "g/kWh", "ELECTRICITY_PV"),
            "HARD_COAL_CGS_70_CHP": PhxCO2Factor(239.9864, "g/kWh", "HARD_COAL_CGS_70_CHP"),
            "HARD_COAL_CGS_35_CHP": PhxCO2Factor(319.9932, "g/kWh", "HARD_COAL_CGS_35_CHP"),
            "HARD_COAL_CGS_0_CHP": PhxCO2Factor(409.9966, "g/kWh", "HARD_COAL_CGS_0_CHP"),
            "GAS_CGS_70_CHP": PhxCO2Factor(-70.0102, "g/kWh", "GAS_CGS_70_CHP"),
            "GAS_CGS_35_CHP": PhxCO2Factor(129.9898, "g/kWh", "GAS_CGS_35_CHP"),
            "GAS_CGS_0_CHP": PhxCO2Factor(319.9932, "g/kWh", "GAS_CGS_0_CHP"),
            "OIL_CGS_70_CHP": PhxCO2Factor(100, "g/kWh", "OIL_CGS_70_CHP"),
            "OIL_CGS_35_CHP": PhxCO2Factor(250.0171, "g/kWh", "OIL_CGS_35_CHP"),
            "OIL_CGS_0_CHP": PhxCO2Factor(409.9966, "g/kWh", "OIL_CGS_0_CHP"),
        }


@dataclass
class PhxLocation:
    """The physical location of the building."""
    latitude: float = 40.6
    longitude: float = -73.8
    site_elevation: Optional[float] = None
    climate_zone: int = 1
    hours_from_UTC: int = -4


@dataclass
class PhxClimatePeakLoad:
    temperature_air: float = 0
    radiation_north: float = 0
    radiation_east: float = 0
    radiation_south: float = 0
    radiation_west: float = 0
    radiation_global: float = 0


@dataclass
class PhxClimate:
    """Monthly Climate Date for the building location."""
    station_elevation: float = 3.0
    selection: SiteClimateSelection = SiteClimateSelection.USER_DEFINED
    daily_temp_swing: float = 8.0
    avg_wind_speed: float = 4.0

    temperature_air: list[float] = field(default_factory=list)
    temperature_dewpoint: list[float] = field(default_factory=list)
    temperature_sky: list[float] = field(default_factory=list)

    radiation_north: list[float] = field(default_factory=list)
    radiation_east: list[float] = field(default_factory=list)
    radiation_south: list[float] = field(default_factory=list)
    radiation_west: list[float] = field(default_factory=list)
    radiation_global: list[float] = field(default_factory=list)

    peak_heating_1: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_heating_2: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_cooling_1: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_cooling_2: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)

@dataclass
class PhxPHPPCodes:
    country_code: str = "US-United States of America"
    region_code: str = "New York"
    dataset_name = "US0055b-New York"
    

@dataclass
class PhxSite:
    """The climate and location date for the building's site."""
    display_name: str = "New York"

    source: str = "__unknown__"
    selection: SiteSelection = SiteSelection.USER_DEFINED
    
    location: PhxLocation = field(default_factory=PhxLocation)
    climate: PhxClimate = field(default_factory=PhxClimate)
    phpp_codes: PhxPHPPCodes = field(default_factory=PhxPHPPCodes)
    ground: PhxGround = field(default_factory=PhxGround)
    
    energy_factors: PhxSiteEnergyFactors = field(default_factory=PhxSiteEnergyFactors)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP Windows worksheet."""

from __future__ import annotations
from typing import List, Optional, Tuple

from PHX.to_PHPP import xl_app, xl_data
from PHX.to_PHPP.xl_data import col_offset
from PHX.to_PHPP.phpp_model import windows_rows
from PHX.to_PHPP.phpp_localization import shape_model


class Windows:
    """IO Controller Class for PHPP "Windows" worksheet."""

    header_row: Optional[int] = None
    first_entry_row: Optional[int] = None
    end_entry_row: Optional[int] = None

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Windows):
        self.xl = _xl
        self.shape = shape

    def find_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.window_rows.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.window_rows.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.window_rows.locator_string_header}?"
        )

    def find_entry_row_start_end(self, _length:int=500) -> Tuple[int, int]:
        """Return the starting and ending rows for the window data entry."""
        start_row = 24
        end_row = 175
        
        if not self.header_row:
            self.header_row = self.find_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.locator_col_entry,
            _row_start=self.header_row,
            _row_end=self.header_row + _length
        )

        for i, val in enumerate(xl_data, start=self.header_row):
            if self.shape.window_rows.locator_string_entry in str(val):
                start_row = i + 2
                break
        else:
            raise Exception(
                f"Error: Cannot find the '{self.shape.window_rows.locator_string_entry}' "
                f"marker on the '{self.shape.name}' sheet, column {self.shape.window_rows.locator_col_entry}?"
            )

        for i, val in enumerate(xl_data, start=self.header_row):
            if self.shape.window_rows_end.locator_string_entry in str(val):
                end_row = i - 3
                break
        else:
            raise Exception(
                f"Error: Cannot find the '{self.shape.window_rows_end.locator_string_entry}' "
                f"marker on the '{self.shape.name}' sheet, column {self.shape.window_rows.locator_col_entry}?"
            )

        return (start_row, end_row)

    def find_section_shape(self) -> None:
        """Find the 'shape' of the Windows worksheet and set the instance attributes."""
        self.header_row = self.find_header_row()
        self.first_entry_row, self.end_entry_row = self.find_entry_row_start_end()

    def write_windows(self, _window_rows: List[windows_rows.WindowRow]) -> None:
        """Write a list of WindowRow objects to the Windows worksheet."""
        if not self.first_entry_row or not self.end_entry_row:
            self.first_entry_row, self.end_entry_row = self.find_entry_row_start_end()

        for i, window_row in enumerate(_window_rows, start=self.first_entry_row):
            for item in window_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def get_all_window_names(self) -> List[str]:
        """Return a list of all the window names found in the worksheet."""
        if not self.first_entry_row or not self.end_entry_row:
            self.first_entry_row, self.end_entry_row = self.find_entry_row_start_end()

        # Get all the window names from the description row
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.inputs.description.column, # type: ignore
            _row_start=self.first_entry_row,
            _row_end=self.end_entry_row
        )

        return [str(_) for _ in xl_data if _]
    
    def activate_variants(self):
        """Set the frame and glass values to link to the Variants worksheet."""
        start_row, end_row = self.find_entry_row_start_end()
        
        for row_num in range(start_row, end_row):
            # -- Link Glazing to the variants type
            self.xl.write_xl_item(
                xl_data.XlItem(
                            self.shape.name,
                            f'{self.shape.window_rows.inputs.glazing_id.column}{row_num}',
                            f'={col_offset(str(self.shape.window_rows.inputs.variant_input.column), 1)}{row_num}',
                        )
            )

            # -- Link Frame to the variants type
            self.xl.write_xl_item(
                xl_data.XlItem(
                            self.shape.name,
                            f'{self.shape.window_rows.inputs.frame_id.column}{row_num}',
                            f'={col_offset(str(self.shape.window_rows.inputs.variant_input.column), 2)}{row_num}',
                        )
            )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Passive House Mechanical Equipment Classes"""

from __future__ import annotations
from inspect import signature
from dataclasses import dataclass, field
from typing import Optional, ClassVar, Union
import uuid

from PHX.model.enums.hvac import DeviceType, SystemType


@dataclass
class PhxUsageProfile:
    """Is the device used to provide..."""
    space_heating: bool = False
    dhw_heating: bool = False
    cooling: bool = False
    ventilation: bool = False
    humidification: bool = False
    dehumidification: bool = False

    def __add__(self, other: PhxUsageProfile) -> PhxUsageProfile:
        obj = self.__class__()
        obj.space_heating = any((self.space_heating, other.space_heating))
        obj.dhw_heating = any((self.dhw_heating, other.dhw_heating))
        obj.cooling = any((self.cooling, other.cooling))
        obj.ventilation = any((self.ventilation, other.ventilation))
        obj.humidification = any((self.humidification, other.humidification))
        obj.dehumidification = any((self.dehumidification, other.dehumidification))
        return obj


@dataclass
class PhxMechanicalDeviceParams:
    """Base class PHX MechanicalEquipment Params"""
    aux_energy: Optional[float] = None
    aux_energy_dhw: Optional[float] = None
    solar_fraction: Optional[float] = None
    in_conditioned_space: bool = True

    @staticmethod
    def safe_add(attr_1, attr_2):
        if not attr_1 and not attr_2:
            return None
        elif not attr_1 and attr_2:
            return attr_2
        elif attr_1 and not attr_2:
            return attr_1
        else:
            return attr_1 + attr_2

    def __add__(self, other: PhxMechanicalDeviceParams) -> PhxMechanicalDeviceParams:
        new_obj = self.__class__()
        new_obj.aux_energy = new_obj.safe_add(self.aux_energy, other.aux_energy)
        new_obj.aux_energy_dhw = new_obj.safe_add(
            self.aux_energy_dhw, other.aux_energy_dhw)
        new_obj.solar_fraction = new_obj.safe_add(
            self.solar_fraction, other.solar_fraction)
        new_obj.in_conditioned_space = any(
            [self.in_conditioned_space, other.in_conditioned_space])
        return new_obj

    def __radd__(self, other):
        if isinstance(other, int):
            return self
        else:
            return self + other


@dataclass
class PhxMechanicalDevice:
    """Base class for PHX Mechanical Devices (heaters, tanks, ventilators)

    This equipment will be part of a PhxMechanicalSubSystem along with distribution.
    """
    _count: ClassVar[int] = 0

    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    id_num: int = field(init=False, default=0)
    system_type: SystemType = SystemType.ANY
    device_type: DeviceType = DeviceType.ELECTRIC
    display_name: str = '_unnamed_equipment_'
    quantity: int = 0
    unit: float = 0.0
    percent_coverage: float = 0.0
    usage_profile: PhxUsageProfile = field(default_factory=PhxUsageProfile)
    params: PhxMechanicalDeviceParams = field(
        default_factory=PhxMechanicalDeviceParams)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count
    
    @property
    def identifier(self):
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str):
        if not _in:
            return
        self._identifier = str(_in)

    def __add__(self, other: PhxMechanicalDevice) -> PhxMechanicalDevice:
        obj = self.__class__()
        obj.device_type = self.device_type
        obj.display_name = self.display_name
        obj.quantity = self.quantity + other.quantity
        obj.unit = self.unit + other.unit
        obj.percent_coverage = self.percent_coverage + other.percent_coverage
        obj.usage_profile = self.usage_profile + other.usage_profile
        obj.params = self.params + other.params
        return obj

    def __radd__(self, other) -> PhxMechanicalDevice:
        if isinstance(other, int):
            return self
        else:
            return self + other

    @classmethod
    def from_kwargs(cls, **kwargs):
        """Allow for the create of base object from arbitrary kwarg input.

        This is used by subclasses during __add__ or __copy__, otherwise fields
        such as 'id_num' or any other init=False fields result in an AttributeError.
        """
        # fetch the constructor's signature
        cls_fields = {field for field in signature(cls).parameters}

        # split the kwargs into native ones and new ones
        native_args, new_args = {}, {}
        for name, val in kwargs.items():
            if name in cls_fields:
                native_args[name] = val
            else:
                new_args[name] = val

        # use the native ones to create the class ...
        return cls(**native_args)

    def __str__(self):
        return f'{self.__class__.__name__}()'


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX  Mechanical Collection Classes."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Optional, List, Any, Union
from collections import defaultdict

from PHX.model import hvac
from PHX.model.enums.hvac import DeviceType
from PHX.model.hvac.heating import AnyPhxHeater
from PHX.model.hvac.ventilation import AnyPhxVentilation
from PHX.model.hvac.cooling import AnyPhxCooling
from PHX.model.hvac.water import AnyWaterTank


class NoVentUnitFoundError(Exception):
    def __init__(self, _id_num):
        self.msg = f"Error: Cannot locate the Mechanical Device with id num: {_id_num}"
        super().__init__(self.msg)


@dataclass
class PhxZoneCoverage:
    """Percentage of the building load-type covered by the subsystem."""
    zone_num: float = 1.0
    heating: float = 1.0
    cooling: float = 1.0
    ventilation: float = 1.0
    humidification: float = 1.0
    dehumidification: float = 1.0

AnyMechDevice = Union[AnyPhxVentilation, AnyPhxHeater, AnyPhxCooling, AnyWaterTank]

@dataclass
class PhxMechanicalSystemCollection:
    """A collection of all the mechanical devices (heating, cooling, etc) and distribution in the project"""
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Ideal Air System"
    sys_type_num: int = 1
    sys_type_str: str = "User defined (ideal system)"
    zone_coverage: PhxZoneCoverage = field(default_factory=PhxZoneCoverage)

    _devices: Dict[str, AnyMechDevice] = field(default_factory=dict)
    _distribution_piping_branches: Dict[str, Any] = field(default_factory=dict)
    _distribution_num_hw_tap_points: int = 1
    _distribution_piping_recirc: Dict[str, Any] = field(default_factory=dict)
    _distribution_ducting: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def devices(self) -> List[AnyMechDevice]:
        return list(self._devices.values())
    
    def device_in_collection(self, _device_key) -> bool:
        """Return True if the a Mech device with the matching key is already in the collection."""
        return _device_key in self._devices.keys()

    def get_mech_device_by_key(self, _key: str) -> Optional[hvac.PhxMechanicalDevice]:
        """Returns the mechanical device with the matching key, or None if not found.

        Arguments:
        ----------
            * _key (str): The key to search the collection for.

        Returns:
        --------
            * (Optional[hvac.PhxMechanicalDevice]) The Mechanical device with
                the matching key, or None if not found.
        """
        return self._devices.get(_key, None)

    def get_mech_device_by_id(self, _id_num: int) -> hvac.PhxMechanicalDevice:
        """Returns a Mechanical Device from the collection which has a matching id-num.

        Arguments:
        ----------
            * _id_num (int): The Mechanical Device id-number to search for.

        Returns:
        --------
            * (hvac.PhxMechanicalDevice): The Mechanical Device found with the
                matching ID-Number. Or Error if not found.
        """
        for device in self._devices.values():
            if device.id_num == _id_num:
                return device

        raise NoVentUnitFoundError(_id_num)

    def add_new_mech_device(self, _key: str, _d: AnyMechDevice) -> None:
        """Adds a new PHX Mechanical device to the collection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new mechanical device
            * _device (_base.PhxMechanicalDevice): The new PHX mechanical device to 
                add to the collection.

        Returns:
        --------
            * None
        """
        self._devices[_key] = _d

    def add_branch_piping(self, _p: hvac.PhxPipeElement) -> None:
        self._distribution_piping_branches[_p.identifier] = _p
    
    def add_recirc_piping(self, _p: hvac.PhxPipeElement) -> None:
        self._distribution_piping_recirc[_p.identifier] = _p

    @property
    def ventilation_devices(self) -> List[hvac.AnyPhxVentilation]:
        """Returns a list of the 'Ventilation' devices in the collection."""
        return [_ for _ in self.devices 
                if isinstance(_, hvac.PhxDeviceVentilation) 
                and _.usage_profile.ventilation
                ]

    @property
    def space_heating_devices(self) -> List[hvac.AnyPhxHeater]:
        """Returns a list of the 'Space Heating' devices in the collection."""
        return [_ for _ in self.devices 
                if isinstance(_, hvac.PhxHeatingDevice) 
                and _.usage_profile.space_heating
                ]

    @property
    def cooling_devices(self) -> List[hvac.AnyPhxCooling]:
        """Returns a list of all the 'Cooling' devices in the collection."""
        return [_ for _ in self.devices 
                if isinstance(_, hvac.PhxCoolingDevice)
                and _.usage_profile.cooling
                ]

    @property
    def dhw_heating_devices(self) -> List[hvac.AnyPhxHeater]:
        """Returns a list of only the 'DHW Heating' devices (no tanks) in the collection."""
        return [_ for _ in self.devices 
                if isinstance(_, hvac.PhxHeatingDevice)
                and _.usage_profile.dhw_heating 
                and _.device_type != DeviceType.WATER_STORAGE
                ]

    @property
    def dhw_tank_devices(self) -> List[hvac.AnyWaterTank]:
        """Returns a list of only the 'DHW Storage Tank' devices (no heaters) in the collection."""
        return [_ for _ in self.devices
                if isinstance(_, hvac.PhxHotWaterTank)
                and _.usage_profile.dhw_heating 
                and _.device_type == DeviceType.WATER_STORAGE
                ]

    @property
    def dhw_branch_piping(self) -> List[hvac.PhxPipeElement]:
        """Returns a list of all the DHW branch-piping in the collection."""
        return list(self._distribution_piping_branches.values())
    
    @property
    def dhw_branch_piping_segments_by_diam(self) -> List[List[hvac.PhxPipeSegment]]:
        """Returns a list of the DHW branch-piping segments, grouped by diameter."""
        # -- Group piping segments by diameter
        d: Dict[float, List[hvac.PhxPipeSegment]] = defaultdict(list)
        for pipe in self.dhw_branch_piping:
            for segment in pipe.segments:
                d[segment.diameter].append(segment)
        
        return list(d.values())

    @property
    def dhw_recirc_piping(self) -> List[hvac.PhxPipeElement]:
        """Returns a list of all the DHW recirculation-piping in the collection."""
        return list(self._distribution_piping_recirc.values())

    @property
    def dhw_recirc_piping_segments_by_diam(self) -> List[List[hvac.PhxPipeSegment]]:
        """Returns a list of the DHW recirculation-piping segments, grouped by diameter."""
        # -- Group piping segments by diameter
        d: Dict[float, List[hvac.PhxPipeSegment]] = defaultdict(list)
        for pipe in self.dhw_recirc_piping:
            for segment in pipe.segments:
                d[segment.diameter].append(segment)
        
        return list(d.values())


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Classes used to build XML Node Objects which are used during XML Output"""

from typing import Union, Collection, Any, Optional

# Type Alias
xml_valid = Union[str, float, int, bool, None]


class XML_Node:
    """A single node text/numeric item. Optional Attribute data"""

    def __init__(
        self, _node_name: str,
        _node_value: xml_valid,
        _attr_name: Optional[str] = None,
        _attr_value: xml_valid = None
    ):
        self.node_name = _node_name
        self.node_value = _node_value
        self.attr_name = _attr_name
        self.attr_value = _attr_value


class XML_List:
    """A List of XML Writable objects. Used to add 'count' info to the list parent node"""

    def __init__(
        self, _node_name: str,
        _node_items: Collection,
        _attr_name: str = "count",
        _attr_value: xml_valid = None
    ):
        self.node_name = _node_name
        self.node_items = _node_items
        self.attr_name = _attr_name
        self._attr_value = _attr_value

    @property
    def attr_value(self):
        if self._attr_value is not None:
            return self._attr_value
        else:
            # -- Used to automate the 'Index' node attr on List items
            return len(self.node_items)

    @attr_value.setter
    def attr_value(self, _in):
        self._attr_value = _in


class XML_Object:
    """XML Writable Object. Object fields will be written out as child nodes"""

    def __init__(
        self,
        _node_name: str,
        _node_object: Any,
        _attr_name: Optional[str] = None,
        _attr_value: xml_valid = None,
        _schema_name: Optional[str] = None,
    ):
        """
        Arguments:
        ----------
            * _node_name (str): The XML name for the node.
            * _node_object (Any): The Honeybee Object to write to the node.
            * _attr_name (Optional[str]): Optional XML node attribute name.
            * _attr_value (Optional[str]): Optional XML node attribute value.
            * _schema_name (Optional[str]): Optional explicit name for xml_schema 
                function to use when writing to XML. If None is passed, uses the
                class name preceded by and underscore.
        """

        self.node_name = _node_name
        self.node_object = _node_object
        self.attr_name = _attr_name
        self.attr_value = _attr_value
        self.schema_name = _schema_name


# For type hints...
xml_writable = Union[XML_Node, XML_List, XML_Object]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "DHW+Distribution" worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP.phpp_model import hot_water_tank, hot_water_piping


class RecircPiping:
    """The Recirculation Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int]=None
    
    @property
    def header_row(self) -> int:
        if not self._header_row: 
            self._header_row = self.find_header_row()
        return self._header_row

    def find_header_row(self, _row_start: int = 100, _rows: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.recirc_piping.locator_col_entry,
            _row_start=_row_start,
            _row_end=_row_start + _rows
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.recirc_piping.locator_string_entry == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.recirc_piping.locator_string_entry}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.recirc_piping.locator_string_entry}?"
        )


class BranchPiping:
    """The Branch Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int]=None

    @property
    def header_row(self) -> int:
        if not self._header_row: 
            self._header_row = self.find_header_row()
        return self._header_row

    def find_header_row(self, _row_start: int = 100, _rows: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.branch_piping.locator_col_entry,
            _row_start=_row_start,
            _row_end=_row_start + _rows
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.branch_piping.locator_string_entry == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.branch_piping.locator_string_entry}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.branch_piping.locator_string_entry}?"
        )


class DHWPiping:
    """The DHW Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape 
        self._header_row: Optional[int] = None
        self.recirc_piping = RecircPiping(self.xl, self.shape)
        self.branch_piping = BranchPiping(self.xl, self.shape)

    def find_header_row(self, _row_start: int = 100, _row_end: int = 200) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.branch_piping.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.branch_piping.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.branch_piping.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.branch_piping.locator_string_header}?"
        )

    @property
    def header_row(self) -> int:
        if not self._header_row: 
            self._header_row = self.find_header_row()
        return self._header_row

    def find_recirc_piping_start_row(self) -> int:
        return self.recirc_piping.header_row

    def find_branch_piping_start_row(self) -> int:
        return self.branch_piping.header_row


class Tank:
    """ An individual tank entry item."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape

    def find_entry_row_start(self, _row_start, num_rows=10) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.tanks.locator_col_entry,
            _row_start=_row_start,
            _row_end=_row_start+num_rows
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.tanks.locator_string_entry == val:
                return i + 2

        raise Exception(
            f"Error: Cannot find the '{self.shape.tanks.locator_string_entry}' "
            f"entry on the '{self.shape.name}' sheet, column {self.shape.tanks.locator_string_entry}?"
        )


class Tanks:
    """The Tanks (Storage Heat Loss) Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape 
        self.header_row: Optional[int] = None
        self.entry_start_row: Optional[int] = None

        self.tank_1 = Tank(self.xl, self.shape)
        self.tank_2 = Tank(self.xl, self.shape)

    def find_header_row(self, _row_start: int = 150, _row_end: int = 200) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.tanks.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.tanks.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.tanks.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.tanks.locator_string_header}?"
        )

    def find_entry_row_start(self) -> int:
        if not self.header_row:
            self.header_row = self.find_header_row()
        return self.tank_1.find_entry_row_start(self.header_row)


class HotWater:
    """IO Controller for the PHPP 'DHW+Distribution' worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self.tanks = Tanks(self.xl, self.shape)
        self.dhw_piping = DHWPiping(self.xl, self.shape)

    def write_tanks(self, _phpp_hw_tanks: List[hot_water_tank.TankInput]) -> None:
        if not self.tanks.entry_start_row:
            self.tanks.entry_start_row = self.tanks.find_entry_row_start()
    
        for phpp_Tank_input in _phpp_hw_tanks:
            for item in phpp_Tank_input.create_xl_items(self.shape.name, self.tanks.entry_start_row):
                self.xl.write_xl_item(item)
    
    def write_branch_piping(self, _phpp_branch_piping: List[hot_water_piping.BranchPipingInput]):
        for pipe_inputs in _phpp_branch_piping:
            for item in pipe_inputs.create_xl_items(self.shape.name, self.dhw_piping.branch_piping.header_row):
                self.xl.write_xl_item(item)
    
    def write_recirc_piping(self, _phpp_recirc_piping: List[hot_water_piping.BranchPipingInput]):
        for pipe_inputs in _phpp_recirc_piping:
            for item in pipe_inputs.create_xl_items(self.shape.name, self.dhw_piping.recirc_piping.header_row):
                self.xl.write_xl_item(item)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Foundation Class"""

from dataclasses import dataclass


@dataclass
class PhxFoundation:
    floor_setting_num: int = 6  # User Defined
    floor_setting_str: str = 'User defined'

    floor_type_num: int = 5
    floor_type_str: str = 'None'



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP DHW Tank"""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model.hvac import water

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model

@dataclass
class TankInput:
    """Model class for a single DHW Tank input."""

    __slots__ = ('shape', 'phx_tank', 'tank_number')
    shape: shape_model.Dhw
    phx_tank: water.PhxHotWaterTank
    tank_number: int

    @property
    def input_column(self) -> str:
        """Return the right input column based on the tank-number."""
        return getattr(self.shape.tanks.input_columns, f"tank_{self.tank_number}")
    
    def create_range(self, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f'{self.input_column}{_row_num}'
    
    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this DHW Tank

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        XLItemDHW = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemDHW(
                self.create_range(_row_num + self.shape.tanks.input_rows.tank_type.row),
                self.shape.tanks.tank_type.options[str(self.phx_tank.params.tank_type.value)]
            ),
            XLItemDHW(
                self.create_range(_row_num + self.shape.tanks.input_rows.standby_losses.row),
                self.phx_tank.params.standby_losses * self.phx_tank.quantity,
                "W/K",
                self.shape.tanks.input_rows.standby_losses.unit
            ),
            XLItemDHW(self.create_range(_row_num + self.shape.tanks.input_rows.storage_capacity.row),
                self.phx_tank.params.storage_capacity * self.phx_tank.quantity,
                "LITER",
                self.shape.tanks.input_rows.storage_capacity.unit
            ),
            XLItemDHW(self.create_range(_row_num + self.shape.tanks.input_rows.standby_fraction.row),
                self.phx_tank.params.standby_fraction
            ),
            XLItemDHW(self.create_range(_row_num + self.shape.tanks.input_rows.tank_location.row),
                self.shape.tanks.tank_location.options[str(int(self.phx_tank.params.in_conditioned_space))]
            ),
            XLItemDHW(self.create_range(_row_num + self.shape.tanks.input_rows.water_temp.row),
                self.phx_tank.params.water_temp,
                "C",
                self.shape.tanks.input_rows.water_temp.unit
            ),
        ]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Areas / Surface-Entry row"""

from dataclasses import dataclass
from typing import List, Optional
from functools import partial

from PHX.model import components, geometry
from PHX.model.enums.building import ComponentExposureExterior, ComponentFaceType

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class SurfaceRow:
    """A single Areas/Surface entry row."""

    __slots__ = ('shape', 'phx_polygon', 'phx_component', 'phpp_assembly_id_name')
    shape: shape_model.Areas
    phx_polygon: geometry.PhxPolygon
    phx_component: components.PhxComponentOpaque
    phpp_assembly_id_name: Optional[str]

    @property
    def phpp_group_number(self) -> int:
        """Return the correct PHPP 'Group Number' depending on the exposure and type."""

        if self.phx_component.exposure_exterior == ComponentExposureExterior.SURFACE:
            return 18
        elif self.phx_component.face_type == ComponentFaceType.WALL:
            if self.phx_component.exposure_exterior == ComponentExposureExterior.EXTERIOR:
                return 8
            else:
                return 9
        elif self.phx_component.face_type == ComponentFaceType.FLOOR:
            return 11
        elif self.phx_component.face_type == ComponentFaceType.ROOF_CEILING:
            return 10
        else:
            return 12

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f'{getattr(self.shape.surface_rows.inputs, _field_name).column}{_row_num}'

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.surface_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAreas = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemAreas(create_range('description'), self.phx_polygon.display_name),
            XLItemAreas(create_range('group_number'), self.phpp_group_number),
            XLItemAreas(create_range('quantity'),  1),
            XLItemAreas(create_range('area'), self.phx_polygon.area, "M2", self._get_target_unit('area')),
            XLItemAreas(create_range('assembly_id'), self.phpp_assembly_id_name),
            XLItemAreas(create_range('orientation'), self.phx_polygon.cardinal_orientation_angle),
            XLItemAreas(create_range('angle'),  self.phx_polygon.angle_from_horizontal),
            XLItemAreas(create_range('shading'),  0.5),
            XLItemAreas(create_range('absorptivity'),  0.6),
            XLItemAreas(create_range('emissivity'),  0.9),
        ]


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP Shading worksheet."""

from __future__ import annotations
from typing import Optional, Tuple, List

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.phpp_model import shading_rows
from PHX.to_PHPP.phpp_localization import shape_model

class Shading:
    """IO Controller Class for PHPP "Shading" worksheet."""

    header_row: Optional[int] = None
    first_entry_row: Optional[int] = None
    end_entry_row: Optional[int] = None

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Shading):
        self.xl = _xl
        self.shape = shape

    def find_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.shading_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.shading_rows.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.shading_rows.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.shading_rows.locator_string_header}?"
        )

    def find_entry_row_start_end(self, _length:int=500) -> Tuple[int, int]:
        """Return the starting and ending rows for the shading data entry."""
        start_row = 17
        end_row = 168
        
        if not self.header_row:
            self.header_row = self.find_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.shading_rows.locator_col_entry,
            _row_start=self.header_row,
            _row_end=self.header_row + _length
        )

        for i, val in enumerate(xl_data, start=self.header_row):
            if self.shape.shading_rows.locator_string_entry in str(val):
                start_row = i + 2
                break
        else:
            raise Exception(
                f"Error: Cannot find the '{self.shape.shading_rows.locator_string_entry}' "
                f"marker on the '{self.shape.name}' sheet, column {self.shape.shading_rows.locator_col_entry}?"
            )

        for i, val in enumerate(xl_data, start=self.header_row):
            if self.shape.shading_rows_end.locator_string_entry in str(val):
                end_row = i - 3
                break
        else:
            raise Exception(
                f"Error: Cannot find the '{self.shape.shading_rows_end.locator_string_entry}' "
                f"marker on the '{self.shape.name}' sheet, column {self.shape.shading_rows.locator_col_entry}?"
            )

        return (start_row, end_row)

    def find_section_shape(self) -> None:
        """Find the 'shape' of the Shading worksheet and set the instance attributes."""
        self.header_row = self.find_header_row()
        self.first_entry_row, self.end_entry_row = self.find_entry_row_start_end()
    
    def write_shading(self, _shading_rows: List[shading_rows.ShadingRow]) -> None:
        """Write a list of ShadingRow objects to the Shading worksheet."""
        if not self.first_entry_row or not self.end_entry_row:
            self.first_entry_row, self.end_entry_row = self.find_entry_row_start_end()

        for i, shading_row in enumerate(_shading_rows, start=self.first_entry_row):
            for item in shading_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to create Project elements from the Honeybee-Model"""

from typing import Optional
from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier
from honeybee import model, room
from PHX.model import project
from PHX.model import schedules
from honeybee_ph_utils.schedules import calc_four_part_vent_sched_values_from_hb_room
from honeybee_energy.schedule import ruleset


def _room_has_ph_style_ventilation(_hb_room: room.Room) -> bool:
    """Returns True if the HB Room has detailed PH-Style ventilation schedule information, False if not.

    Arguments:
    ---------
        * _hb_room (room.Room): The Honeybee Room to look at.
    Returns:
    --------
        * bool:
    """

    if not _hb_room.properties.energy:
        return False
    elif _hb_room.properties.energy.ventilation.schedule is None:
        return False
    elif not _hb_room.properties.energy.ventilation.schedule.properties.ph.daily_operating_periods:
        return False
    else:
        return True


def _get_room_vent_pattern_key(_hb_room: room.Room) -> str:
    """Returns the right 'key' for the Honeybee Room's ventilation pattern.

    When translating HB-Style ventilation over, each HB-Room will get a different 
    utilization-pattern key since the schedule-periods will be dependant on the rooms' occupancy 
    and size.

    Is there a way to 'group' HB-Style ventilation patterns ahead of time? 

    If there is detailed PH-Style ventilation schedule info, that can be used instead 
    of the HB-Style.

    Arguments:
    ---------
        * _hb_room (room.Room): The Honeybee Room to look at.
    Returns:
    --------
        * str: The ventilation pattern 'key'
    """

    if _room_has_ph_style_ventilation(_hb_room):
        return _hb_room.properties.energy.ventilation.schedule.identifier
    else:
        # Room ventilation pattern needs to be created from the HB info and
        # take into account occupancy, size, etc...
        return _hb_room.properties.energy.ventilation.identifier


def _create_vent_pattern_from_hb_style(_hb_room: room.Room) -> schedules.UtilizationPatternVent:
    """Returns a new PHX Utilization Pattern (Vent) based on the HB-Room's E+ 
    HB-Style info found on the energy.ventilation

    This is used when no detailed PH-Style info is set by the user and you want to 
    convert over an existing E+/HB style ventilation object to PH-Style.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the new PHX UtilizationPatternVent from.

    Returns:
    --------
        * schedules.UtilizationPatternVent: A new PHX Util Pattern.
    """

    new_util_pattern = schedules.UtilizationPatternVent()

    wufi_sched = calc_four_part_vent_sched_values_from_hb_room(_hb_room)

    new_util_pattern.operating_periods.high.period_operating_hours = wufi_sched.high.period_operating_hours
    new_util_pattern.operating_periods.high.period_operation_speed = wufi_sched.high.period_speed
    new_util_pattern.operating_periods.standard.period_operating_hours = wufi_sched.standard.period_operating_hours
    new_util_pattern.operating_periods.standard.period_operation_speed = wufi_sched.standard.period_speed
    new_util_pattern.operating_periods.basic.period_operating_hours = wufi_sched.basic.period_operating_hours
    new_util_pattern.operating_periods.basic.period_operation_speed = wufi_sched.basic.period_speed
    new_util_pattern.operating_periods.minimum.period_operating_hours = wufi_sched.minimum.period_operating_hours
    new_util_pattern.operating_periods.minimum.period_operation_speed = wufi_sched.minimum.period_speed

    # -- Keep all the IDs in alignment....
    new_util_pattern.identifier = _hb_room.properties.energy.ventilation.identifier
    new_util_pattern.id_num = new_util_pattern._count
    _hb_room.properties.energy.ventilation.schedule.properties.ph.id_num = new_util_pattern.id_num  # <--- Important!
    new_util_pattern.name = _hb_room.properties.energy.ventilation.display_name

    return new_util_pattern


def _create_vent_pattern_from_ph_style(_hb_room: room.Room) -> schedules.UtilizationPatternVent:
    """Returns a new PHX Utilization Pattern (Vent) based on the HB-Room's detailed 
    PH-Style info found on the energy.ventilation.schedule.properties.ph

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to build the new PHX UtilizationPatternVent from.

    Returns:
    --------
        * schedules.UtilizationPatternVent: A new PHX Util Pattern.
    """

    new_util_pattern = schedules.UtilizationPatternVent()

    ph_sched_props = _hb_room.properties.energy.ventilation.schedule.properties.ph

    new_util_pattern.operating_hours = 24.0
    new_util_pattern.operating_days = ph_sched_props.operating_days_wk
    new_util_pattern.operating_weeks = ph_sched_props.operating_weeks_year

    for op_period in ph_sched_props.daily_operating_periods:
        phx_vent_util_period = getattr(
            new_util_pattern.operating_periods, op_period.name)
        phx_vent_util_period.period_operating_hours = op_period.operation_hours
        phx_vent_util_period.period_operation_speed = op_period.operation_fraction
        setattr(new_util_pattern.operating_periods, op_period.name, phx_vent_util_period)

    # -- Keep all the IDs in alignment....
    new_util_pattern.identifier = _hb_room.properties.energy.ventilation.schedule.identifier
    new_util_pattern.id_num = new_util_pattern._count
    _hb_room.properties.energy.ventilation.schedule.properties.ph.id_num = new_util_pattern.id_num  # <--- Important!
    new_util_pattern.name = _hb_room.properties.energy.ventilation.schedule.display_name

    return new_util_pattern


def build_util_pat_from_hb_room(_hb_room: room.Room) -> Optional[schedules.UtilizationPatternVent]:
    """Build a new Ventilation Utilization Schedule based on a Honeybee-Room's energy.ventilation values.

    Arguments:
    ----------
        *_hb_room (room.Room): The Honeybee-Room to get the ventilation pattern data from.

    Returns:
    --------
        * schedules.UtilizationPatternVent | None: The new Ventilation Pattern or None if no
            energy.ventilation or energy.ventilation.schedule is found on the room.
    """

    if _hb_room.properties.energy.ventilation is None:
        return None
    elif _hb_room.properties.energy.ventilation.schedule is None:
        return None
    elif _room_has_ph_style_ventilation(_hb_room):
        # -- The room's ventilation DOES have detailed user-inputs for operation periods, so use those.
        # -- This is the case when a Honeybee-PH component is used to build the PH_ScheduleRuleset
        new_util_pattern = _create_vent_pattern_from_ph_style(_hb_room)
    else:
        # -- There IS a ventilation.schedule, BUT there are no detailed
        # -- Passive House style user-input operation periods set on it. This is the case when a normal
        # -- HB-Ventilation ScheduleRuleset is converted over to a PH_ScheduleRuleset
        new_util_pattern = _create_vent_pattern_from_hb_style(_hb_room)

    # -- Ensure that the operating hours add up to exactly 24
    new_util_pattern.force_max_utilization_hours()

    return new_util_pattern


def _add_default_vent_schedule_to_Rooms(_hb_model: model.Model) -> model.Model:
    """Add a default ventilation.schedule to the HB Model's Rooms if they have None. 

    Some HB Programs do not have a ventilation.schedule. I *think* this means 
    constant operation. So add the constant value (1) default ventilation schedule to the room.

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to add the new constant value
            ventilation schedules to.

    Returns:
    --------
        * model.Model: The Honeybee Model with new ventilation schedules added to 
            any Rooms which were missing them.
    """

    type_limit = schedule_type_limit_by_identifier('Fractional')
    default_ventilation_schedule = ruleset.ScheduleRuleset.from_constant_value(
        'default_schedule', 1.0, type_limit)

    for hb_room in _hb_model.rooms:
        if hb_room.properties.energy.ventilation.schedule is None:
            hb_room.properties.energy.ventilation.unlock()
            hb_room.properties.energy.ventilation.schedule = default_ventilation_schedule
            hb_room.properties.energy.ventilation.lock()

    return _hb_model


def build_util_patterns_ventilation_from_HB_Model(_project: project.PhxProject, _hb_model: model.Model) -> None:
    """Build the Utilization Pattern collection for Ventilation patterns and add to the PHX-Project.

    Arguments:
    ----------
        * _project (project.Project): The PHX-Project to add the new Utilization Patterns to.
        * _hb_model (model.Model): Then Honeybee Model to build up the new Utilization Patterns from.

    Returns:
    --------
        * None
    """

    # -- FIRST: Have to clean up the HB-ventilation schedules where they are missing.
    _hb_model = _add_default_vent_schedule_to_Rooms(_hb_model)

    # -- NEXT: Build up the new Ventilation Patterns from the Room's data
    for hb_room in _hb_model.rooms:
        vent_pattern_id = _get_room_vent_pattern_key(hb_room)
        if _project.utilization_patterns_ventilation.key_is_in_collection(vent_pattern_id):
            # -- This is just to help speed things up.
            # -- Don't re-make the util-pattern if it is already in collection.
            continue

        new_util_pattern = build_util_pat_from_hb_room(hb_room)
        _project.utilization_patterns_ventilation.add_new_util_pattern(new_util_pattern)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for the Ventilation worksheet various input items."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.xl_data import xl_writable
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class VentilationInputItem:
    """Ventilation Worksheet input item."""

    shape: shape_model.Ventilation
    input_data: xl_writable
    input_type: str = field(init=False)
    input_unit: Optional[str] = None
    target_unit: Optional[str] = None

    def create_xl_item(self, _sheet_name: str, _row_num: int) -> xl_data.XlItem:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for

        Returns:
        --------
            * (XlItem): The XlItem to write to the sheet.
        """
        return xl_data.XlItem(
            sheet_name=_sheet_name,
            xl_range=f'{getattr(self.shape, self.input_type).input_column}{_row_num}',
            write_value=self.input_data,
            input_unit=self.input_unit,
            target_unit=self.target_unit
        )

    @classmethod
    def vent_type(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'vent_type'
        return obj

    @classmethod
    def multi_unit_on(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'multi_unit_on'
        return obj

    @classmethod
    def wind_coeff_e(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'wind_coeff_e'
        return obj

    @classmethod
    def wind_coeff_f(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'wind_coeff_f'
        return obj

    @classmethod
    def airtightness_n50(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'airtightness_n50'
        return obj

    @classmethod
    def airtightness_Vn50(cls, shape: shape_model.Ventilation, input_data: xl_writable) -> VentilationInputItem:
        obj = cls(shape, input_data)
        obj.input_type = 'airtightness_Vn50'
        obj.input_unit = "M3"
        obj.target_unit = shape.airtightness_Vn50.unit
        return obj



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Conversion Schemas for how to write PH/HB objects to WUFI XML"""

from typing import List, Optional
import sys

from PHX.model import (certification, constructions,
                       geometry, ground, phx_site, schedules, hvac, loads,
                       building, elec_equip, project, components)
from PHX.to_WUFI_XML.xml_writables import XML_Node, XML_List, XML_Object, xml_writable

TOL_LEV1 = 2  # Value tolerance for rounding. ie; 9.843181919194 -> 9.84
TOL_LEV2 = 10  # Value tolerance for rounding. ie; 9.843181919194 -> 9.8431819192

# -- PROJECT ------------------------------------------------------------------


def _PhxProject(_wufi_project: project.PhxProject) -> List[xml_writable]:
    return [
        XML_Node("DataVersion", _wufi_project.data_version),
        XML_Node("UnitSystem", _wufi_project.data_version),
        XML_Node("ProgramVersion", _wufi_project.program_version),
        XML_Node("Scope", _wufi_project.scope),
        XML_Node("DimensionsVisualizedGeometry",
                 _wufi_project.visualized_geometry),
        XML_Object("ProjectData", _wufi_project.project_data),
        XML_List("UtilisationPatternsVentilation",
                 [XML_Object("UtilizationPatternVent", pat, "index", i)
                  for i, pat in enumerate(_wufi_project.utilization_patterns_ventilation)]),
        XML_List("UtilizationPatternVent",
                 _wufi_project.utilization_patterns_ph),
        XML_List("Variants", [XML_Object("Variant", var, "index", i)
                 for i, var in enumerate(_wufi_project.variants)]),
        XML_List(
            "Assemblies",
            [XML_Object("Assembly", at_id, "index", i)
             for i, at_id in enumerate(_wufi_project.assembly_types.values())],
        ),
        XML_List(
            "WindowTypes",
            [XML_Object("WindowType", wt_id, "index", i)
             for i, wt_id in enumerate(_wufi_project.window_types.values())],
        ),
    ]


def _PhxVariant(_variant: project.PhxVariant) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _variant.id_num),
        XML_Node("Name", _variant.name),
        XML_Node("Remarks", _variant.remarks),
        XML_Node("PlugIn", _variant.plugin),
        XML_Object("Graphics_3D", _variant.graphics3D),
        XML_Object("Building", _variant.building),
        XML_Object("ClimateLocation", _variant.site),
        XML_Object("PassivehouseData", _variant.phius_certification),
        XML_Object("HVAC", _variant.mech_systems,
                   _schema_name='_PhxMechanicalSystemCollection'),
    ]


def _PhxProjectData(_project_data: project.PhxProjectData) -> List[xml_writable]:
    return [
        XML_Node("Year_Construction", _project_data.year_constructed),
        XML_Node("OwnerIsClient", _project_data.owner_is_client),
        XML_Node("Date_Project", _project_data.project_date),
        XML_Node("WhiteBackgroundPictureBuilding", _project_data.image),
        XML_Node("Customer_Name", _project_data.customer.name),
        XML_Node("Customer_Street", _project_data.customer.street),
        XML_Node("Customer_Locality", _project_data.customer.city),
        XML_Node("Customer_PostalCode", _project_data.customer.post_code),
        XML_Node("Customer_Tel", _project_data.customer.telephone),
        XML_Node("Customer_Email", _project_data.customer.email),
        XML_Node("Building_Name", _project_data.building.name),
        XML_Node("Building_Street", _project_data.building.street),
        XML_Node("Building_Locality", _project_data.building.city),
        XML_Node("Building_PostalCode", _project_data.building.post_code),
        XML_Node("Owner_Name", _project_data.owner.name),
        XML_Node("Owner_Street", _project_data.owner.street),
        XML_Node("Owner_Locality", _project_data.owner.city),
        XML_Node("Owner_PostalCode", _project_data.owner.post_code),
        XML_Node("Responsible_Name", _project_data.responsible.name),
        XML_Node("Responsible_Street", _project_data.responsible.street),
        XML_Node("Responsible_Locality", _project_data.responsible.city),
        XML_Node("Responsible_PostalCode",
                 _project_data.responsible.post_code),
        XML_Node("Responsible_Tel", _project_data.responsible.telephone),
        XML_Node("Responsible_LicenseNr",
                 _project_data.responsible.license_number),
        XML_Node("Responsible_Email", _project_data.responsible.email),
    ]


# -- BUILDING -----------------------------------------------------------------


def _PhxBuilding(_b: building.PhxBuilding) -> List[xml_writable]:
    return [
        XML_List("Components", [XML_Object("Component", c, "index", i)
                 for i, c in enumerate(_b.all_components)]),
        XML_List("Zones", [XML_Object("Zone", z, "index", i)
                 for i, z in enumerate(_b.zones)]),
    ]


def _PhxZone(_z: building.PhxZone) -> List[xml_writable]:
    return [
        XML_Node("Name", _z.display_name),
        XML_Node("KindZone", 1, "choice", "Simulated zone"),
        XML_Node("IdentNr", _z.id_num),
        XML_List("RoomsVentilation", [XML_Object("Room", rm, "index", i, _schema_name="_PhxRoomVentilation")
                                      for i, rm in enumerate(_z.wufi_rooms)]),
        XML_Node("GrossVolume_Selection", 6),
        XML_Node("GrossVolume", _z.volume_gross),
        XML_Node("NetVolume_Selection", 6),
        XML_Node("NetVolume", _z.volume_net),
        XML_Node("FloorArea_Selection", 6),
        XML_Node("FloorArea", _z.weighted_net_floor_area),
        XML_Node("ClearanceHeight_Selection", 1),
        XML_Node("ClearanceHeight", _z.clearance_height),
        XML_Node("SpecificHeatCapacity_Selection", 2),
        XML_Node("SpecificHeatCapacity", _z.specific_heat_capacity),
        XML_Node("IdentNrPH_Building", 1),
        XML_Node("OccupantQuantityUserDef", int(
            _z.res_occupant_quantity), "unit", "-"),
        XML_Node("NumberBedrooms", int(_z.res_number_bedrooms), "unit", "-"),
        XML_List("HomeDevice", [XML_Object("Device", d, "index", i, _schema_name='_PhxElectricalDevice')
                 for i, d in enumerate(_z.elec_equipment_collection.devices)]),
    ]


def _PhxComponentOpaque(_c: components.PhxComponentOpaque) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _c.id_num),
        XML_Node("Name", _c.display_name),
        XML_Node("Visual", True),
        XML_Node("Type", _c.face_opacity.value),
        XML_Node("IdentNrColorI", _c.color_interior.value),
        XML_Node("IdentNrColorE", _c.color_exterior.value),
        XML_Node("InnerAttachment", _c.exposure_interior),
        XML_Node("OuterAttachment", _c.exposure_exterior.value),
        XML_Node("IdentNr_ComponentInnerSurface", _c.interior_attachment_id),
        XML_Node("IdentNrAssembly", _c.assembly_type_id_num),
        XML_Node("IdentNrWindowType", -1),
        XML_List("IdentNrPolygons", [XML_Node(
            "IdentNr", n, "index", i) for i, n in enumerate(_c.polygon_ids)]),
    ]


def _PhxComponentAperture(_c: building.PhxComponentAperture) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _c.id_num),
        XML_Node("Name", _c.display_name),
        XML_Node("Visual", True),
        XML_Node("Type", _c.face_opacity.value),
        XML_Node("IdentNrColorI", _c.color_interior.value),
        XML_Node("IdentNrColorE", _c.color_exterior.value),
        XML_Node("InnerAttachment", _c.exposure_interior),
        XML_Node("OuterAttachment", _c.exposure_exterior.value),
        XML_Node("IdentNr_ComponentInnerSurface", _c.interior_attachment_id),
        XML_Node("IdentNrAssembly", -1),
        XML_Node("IdentNrWindowType", _c.window_type_id_num),
        XML_List("IdentNrPolygons", [XML_Node(
            "IdentNr", n, "index", i) for i, n in enumerate(_c.polygon_ids)]),
    ]


# -- CERTIFICATION ------------------------------------------------------------


def _PhxPhBuildingData(_phius_cert: certification.PhxPhiusCertification) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _phius_cert.ph_building_data._count),
        XML_Node("BuildingCategory", _phius_cert.phius_certification_settings.phius_building_category_type.value),
        XML_Node("OccupancyTypeResidential", _phius_cert.phius_certification_settings.phius_building_use_type.value),
        XML_Node("BuildingStatus", _phius_cert.phius_certification_settings.phius_building_status.value),
        XML_Node("BuildingType", _phius_cert.phius_certification_settings.phius_building_type.value),
        XML_Node("OccupancySettingMethod", _phius_cert.ph_building_data.occupancy_setting_method),
        XML_Node("NumberUnits", _phius_cert.ph_building_data.num_of_units),
        XML_Node("CountStories", _phius_cert.ph_building_data.num_of_floors),
        XML_Node("EnvelopeAirtightnessCoefficient", _phius_cert.ph_building_data.airtightness_q50),
        XML_List('FoundationInterfaces', [XML_Object('FoundationInterface', f, 'index', i)
                                          for i, f in enumerate(_phius_cert.ph_building_data.foundations)]),
    ]


def _PhxPhiusCertification(_phius_cert: certification.PhxPhiusCertification) -> List[xml_writable]:
    # No idea why this is a list in Wufi? When would there ever be more than 1? whatever...
    _temp_bldg_data_list = [_phius_cert]
    
    return [
        XML_Node("PH_CertificateCriteria",
                 _phius_cert.phius_certification_criteria.ph_certificate_criteria),
        XML_Node("PH_SelectionTargetData",
                 _phius_cert.phius_certification_criteria.ph_selection_target_data),
        XML_Node("AnnualHeatingDemand",
                 _phius_cert.phius_certification_criteria.phius_annual_heating_demand),
        XML_Node("AnnualCoolingDemand",
                 _phius_cert.phius_certification_criteria.phius_annual_cooling_demand),
        XML_Node("PeakHeatingLoad",
                 _phius_cert.phius_certification_criteria.phius_peak_heating_load),
        XML_Node("PeakCoolingLoad",
                 _phius_cert.phius_certification_criteria.phius_peak_cooling_load),
        XML_List("PH_Buildings", [XML_Object("PH_Building", obj, "index", i, _schema_name="_PhxPhBuildingData")
                 for i, obj in enumerate(_temp_bldg_data_list)]),
    ]


# -- FOUNDATIONS --------------------------------------------------------------

def _PhxFoundation(_f: ground.PhxFoundation) -> List[xml_writable]:
    return [
        XML_Node("Name", ''),
        XML_Node("SettingFloorSlabType", _f.floor_setting_num,
                 'choice', _f.floor_setting_str),
        XML_Node("FloorSlabType", _f.floor_type_num,
                 'choice', _f.floor_type_str),

        # XML_Node("PositionPerimeterInsulation", 1),
        # XML_Node("PerimeterInsulationWidthDepth", 1),
        # XML_Node("ThicknessPerimeterInsulation", 1),
        # XML_Node("ConductivityPerimeterInsulation", 1),
        # XML_Node("PhaseShiftMonths", 1),
        # XML_Node("HarmonicFraction", 1),
        # XML_Node("BasementVentilationACH", 1),
        # XML_Node("DepthBasementBelowGroundSurface_Selection", 1),
        # XML_Node("DepthBasementBelowGroundSurface", 1),
        # XML_Node("HeightBasementWallAboveGrade_Selection", 1),
        # XML_Node("HeightBasementWallAboveGrade", 1),
        # XML_Node("CrawlspaceVentOpenings_Selection", 1),
        # XML_Node("CrawlspaceVentOpenings", 1),
        # XML_Node("FloorSlabArea_Selection", 1),
        # XML_Node("FloorSlabArea", 1),
        # XML_Node("U_ValueBasementSlab_Selection", 1),
        # XML_Node("U_ValueBasementSlab", 1),
        # XML_Node("FloorCeilingArea_Selection", 1),
        # XML_Node("FloorCeilingArea", 1),
        # XML_Node("U_ValueCeilingToUnheatedCellar_Selection", 1),
        # XML_Node("U_ValueCeilingToUnheatedCellar", 1),
        # XML_Node("U_ValueBasementWall_Selection", 1),
        # XML_Node("U_ValueBasementWall", 1),
        # XML_Node("U_ValueWallAboveGround_Selection", 1),
        # XML_Node("U_ValueWallAboveGround", 1),
        # XML_Node("FloorSlabPerimeter_Selection", 1),
        # XML_Node("FloorSlabPerimeter", 1),
        # XML_Node("BasementVolume_Selection", 1),
        # XML_Node("BasementVolume", 1),
        # XML_Node("U_ValueCrawlspaceFloor_Selection", 1),
        # XML_Node("U_ValueCrawlspaceFloor", 1),
    ]


# -- CLIMATE ------------------------------------------------------------------


def _PH_ClimateLocation(_phx_site: phx_site.PhxSite) -> List[xml_writable]:

    def _in_wufi_order(_factor_dict: dict) -> List[phx_site.PhxEnergyFactorAlias]:
        """Returns the PE /CO2 conversion factors in WUFI-specific order."""
        fuel_order = ["OIL", "NATURAL_GAS", "LPG", "HARD_COAL", "WOOD", "ELECTRICITY_MIX",
                      "ELECTRICITY_PV", "HARD_COAL_CGS_70_CHP", "HARD_COAL_CGS_35_CHP",
                      "HARD_COAL_CGS_0_CHP", "GAS_CGS_70_CHP", "GAS_CGS_35_CHP",
                      "GAS_CGS_0_CHP", "OIL_CGS_70_CHP", "OIL_CGS_35_CHP", "OIL_CGS_0_CHP",
                      ]
        return [_factor_dict[fuel_name] for fuel_name in fuel_order]

    return [
        XML_Node('Selection', _phx_site.climate.selection.value),
        XML_Node('DailyTemperatureSwingSummer',
                 _phx_site.climate.daily_temp_swing),
        XML_Node('AverageWindSpeed', _phx_site.climate.avg_wind_speed),

        # -- Location
        XML_Node('Latitude', _phx_site.location.latitude),
        XML_Node('Longitude', _phx_site.location.longitude),
        XML_Node('HeightNNWeatherStation',
                 _phx_site.climate.station_elevation),
        XML_Node('dUTC', _phx_site.location.hours_from_UTC),
        XML_Node('ClimateZone', _phx_site.location.climate_zone),

        # -- Ground
        XML_Node('GroundThermalConductivity',
                 _phx_site.ground.ground_thermal_conductivity),
        XML_Node('GroundHeatCapacitiy',
                 _phx_site.ground.ground_heat_capacity),
        XML_Node('GroundDensity', _phx_site.ground.ground_density),
        XML_Node('DepthGroundwater', _phx_site.ground.depth_groundwater),
        XML_Node('FlowRateGroundwater',
                 _phx_site.ground.flow_rate_groundwater),

        # -- Monthly
        XML_List('TemperatureMonthly', [XML_Node("Item", val, "index", i)
                                        for i, val in enumerate(_phx_site.climate.temperature_air)]),
        XML_List('DewPointTemperatureMonthly', [XML_Node("Item", val, "index", i)
                                                for i, val in enumerate(_phx_site.climate.temperature_dewpoint)]),
        XML_List('SkyTemperatureMonthly', [XML_Node("Item", val, "index", i)
                                           for i, val in enumerate(_phx_site.climate.temperature_sky)]),
        XML_List('NorthSolarRadiationMonthly', [XML_Node("Item", val, "index", i)
                                                for i, val in enumerate(_phx_site.climate.radiation_north)]),
        XML_List('EastSolarRadiationMonthly', [XML_Node("Item", val, "index", i)
                                               for i, val in enumerate(_phx_site.climate.radiation_east)]),
        XML_List('SouthSolarRadiationMonthly', [XML_Node("Item", val, "index", i)
                                                for i, val in enumerate(_phx_site.climate.radiation_south)]),
        XML_List('WestSolarRadiationMonthly', [XML_Node("Item", val, "index", i)
                                               for i, val in enumerate(_phx_site.climate.radiation_west)]),
        XML_List('GlobalSolarRadiationMonthly', [XML_Node("Item", val, "index", i)
                                                 for i, val in enumerate(_phx_site.climate.radiation_global)]),

        # -- Peak Load Values
        XML_Node('TemperatureHeating1',
                 _phx_site.climate.peak_heating_1.temperature_air),
        XML_Node('NorthSolarRadiationHeating1',
                 _phx_site.climate.peak_heating_1.radiation_north),
        XML_Node('EastSolarRadiationHeating1',
                 _phx_site.climate.peak_heating_1.radiation_east),
        XML_Node('SouthSolarRadiationHeating1',
                 _phx_site.climate.peak_heating_1.radiation_south),
        XML_Node('WestSolarRadiationHeating1',
                 _phx_site.climate.peak_heating_1.radiation_west),
        XML_Node('GlobalSolarRadiationHeating1',
                 _phx_site.climate.peak_heating_1.radiation_global),

        XML_Node('TemperatureHeating2',
                 _phx_site.climate.peak_heating_2.temperature_air),
        XML_Node('NorthSolarRadiationHeating2',
                 _phx_site.climate.peak_heating_2.radiation_north),
        XML_Node('EastSolarRadiationHeating2',
                 _phx_site.climate.peak_heating_2.radiation_east),
        XML_Node('SouthSolarRadiationHeating2',
                 _phx_site.climate.peak_heating_2.radiation_south),
        XML_Node('WestSolarRadiationHeating2',
                 _phx_site.climate.peak_heating_2.radiation_west),
        XML_Node('GlobalSolarRadiationHeating2',
                 _phx_site.climate.peak_heating_2.radiation_global),

        XML_Node('TemperatureCooling',
                 _phx_site.climate.peak_cooling_1.temperature_air),
        XML_Node('NorthSolarRadiationCooling',
                 _phx_site.climate.peak_cooling_1.radiation_north),
        XML_Node('EastSolarRadiationCooling',
                 _phx_site.climate.peak_cooling_1.radiation_east),
        XML_Node('SouthSolarRadiationCooling',
                 _phx_site.climate.peak_cooling_1.radiation_south),
        XML_Node('WestSolarRadiationCooling',
                 _phx_site.climate.peak_cooling_1.radiation_west),
        XML_Node('GlobalSolarRadiationCooling',
                 _phx_site.climate.peak_cooling_1.radiation_global),

        XML_Node('TemperatureCooling2',
                 _phx_site.climate.peak_cooling_2.temperature_air),
        XML_Node('NorthSolarRadiationCooling2',
                 _phx_site.climate.peak_cooling_2.radiation_north),
        XML_Node('EastSolarRadiationCooling2',
                 _phx_site.climate.peak_cooling_2.radiation_east),
        XML_Node('SouthSolarRadiationCooling2',
                 _phx_site.climate.peak_cooling_2.radiation_south),
        XML_Node('WestSolarRadiationCooling2',
                 _phx_site.climate.peak_cooling_2.radiation_west),
        XML_Node('GlobalSolarRadiationCooling2',
                 _phx_site.climate.peak_cooling_2.radiation_global),

        XML_Node('SelectionPECO2Factor',
                 _phx_site.energy_factors.selection_pe_co2_factor.value),
        XML_List('PEFactorsUserDef', [XML_Node(f"PEF{i}", factor.value, "unit", factor.unit)
                                      for i, factor in enumerate(_in_wufi_order(_phx_site.energy_factors.pe_factors))]),
        XML_List('CO2FactorsUserDef', [XML_Node(f"CO2F{i}", factor.value, "unit", factor.unit)
                                       for i, factor in enumerate(_in_wufi_order(_phx_site.energy_factors.co2_factors))]),
    ]


def _PhxSite(_phx_site: phx_site.PhxSite) -> List[xml_writable]:
    return [
        XML_Node('Selection', _phx_site.selection.value),
        # XML_Node('IDNr_DB', _climate.),
        # XML_Node('Name_DB', _climate.),
        # XML_Node('Comment_DB', _climate.),
        XML_Node('Latitude_DB', _phx_site.location.latitude, 'unit', "°"),
        XML_Node('Longitude_DB', _phx_site.location.longitude,  'unit', "°"),
        XML_Node(
            'HeightNN_DB', _phx_site.location.site_elevation, 'unit', "m"),
        XML_Node('dUTC_DB', _phx_site.location.hours_from_UTC),
        # XML_Node('FileName_DB', _climate.),
        # XML_Node('Type_DB', _climate.),
        # XML_Node('CatalogueNr_DB', _climate.),
        # XML_Node('MapNr_DB', _climate.),
        XML_Node('Albedo', -2, 'choice', "User defined"),
        XML_Node('GroundReflShort', 0.2, 'unit', "-"),
        XML_Node('GroundReflLong', 0.1, 'unit', "-"),
        XML_Node('GroundEmission', 0.9, 'unit', "-"),
        XML_Node('CloudIndex', 0.66, 'unit', "-"),
        XML_Node('CO2concenration', 350, 'unit', "mg/m³"),
        XML_Node('Unit_CO2concentration', 48, 'choice', "ppmv"),
        XML_Object('PH_ClimateLocation', _phx_site,
                   _schema_name='_PH_ClimateLocation'),
    ]


# -- GEOMETRY -----------------------------------------------------------------


def _PhxGraphics3D(_graphics3D: geometry.PhxGraphics3D) -> List[xml_writable]:
    return [
        XML_List("Vertices", [XML_Object("Vertix", var, "index", i)
                 for i, var in enumerate(_graphics3D.vertices)]),
        XML_List("Polygons", [XML_Object("Polygon", var, "index", i, _schema_name="_PhxPolygon")
                 for i, var in enumerate(_graphics3D.polygons)]),
    ]


def _PhxPolygon(_p: geometry.PhxPolygon) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _p.id_num),
        XML_Node("NormalVectorX", round(_p.normal_vector.x, TOL_LEV2)),
        XML_Node("NormalVectorY", round(_p.normal_vector.y, TOL_LEV2)),
        XML_Node("NormalVectorZ", round(_p.normal_vector.z, TOL_LEV2)),
        XML_List(
            "IdentNrPoints", [XML_Node("IdentNr", v_id, "index", i)
                              for i, v_id in enumerate(_p.vertices_id_numbers)]
        ),
        XML_List(
            "IdentNrPolygonsInside",
            [XML_Node("IdentNr", v_id, "index", i)
             for i, v_id in enumerate(_p.child_polygon_ids)],
        ),
    ]


def _PhxVertix(_v: geometry.PhxVertix) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _v.id_num),
        XML_Node("X", round(_v.x, TOL_LEV2)),
        XML_Node("Y", round(_v.y, TOL_LEV2)),
        XML_Node("Z", round(_v.z, TOL_LEV2)),
    ]


# -- CONSTRUCTIONS ------------------------------------------------------------


def _PhxConstructionOpaque(_a: constructions.PhxConstructionOpaque) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _a.id_num),
        XML_Node("Name", _a.display_name),
        XML_Node("Order_Layers", _a.layer_order),
        XML_Node("Grid_Kind", _a.grid_kind),
        XML_List("Layers", [XML_Object("Layer", n, "index", i)
                 for i, n in enumerate(_a.layers)]),
    ]


def _PhxLayer(_l: constructions.PhxLayer) -> List[xml_writable]:
    return [
        XML_Node("Thickness", _l.thickness_m),
        XML_Object("Material", _l.material),
    ]


def _PhxMaterial(_m: constructions.PhxMaterial) -> List[xml_writable]:
    return [
        XML_Node("Mass", _m.display_name),
        XML_Node("ThermalConductivity", _m.conductivity),
        XML_Node("BulkDensity", _m.density),
        XML_Node("Porosity", _m.porosity),
        XML_Node("HeatCapacity", _m.heat_capacity),
        XML_Node("WaterVaporResistance", _m.water_vapor_resistance),
        XML_Node("ReferenceW", _m.reference_water),
    ]


def _PhxConstructionWindow(_wt: constructions.PhxConstructionWindow) -> List[xml_writable]:
    return [
        XML_Node("IdentNr", _wt.id_num),
        XML_Node("Name", _wt.display_name),
        XML_Node("Uw_Detailed>", _wt.use_detailed_uw),
        XML_Node("GlazingFrameDetailed", _wt.use_detailed_frame),
        XML_Node("FrameFactor", _wt.frame_factor),
        XML_Node("U_Value", _wt.u_value_window),
        XML_Node("U_Value_Glazing", _wt.u_value_glass),
        XML_Node("MeanEmissivity", _wt.glass_mean_emissivity),
        XML_Node("g_Value", _wt.glass_g_value),
        XML_Node("SHGC_Hemispherical", _wt.glass_g_value),
        # --
        XML_Node("U_Value_Frame", _wt.u_value_frame),
        # --
        XML_Node("Frame_Width_Left", _wt.frame_left.width),
        XML_Node("Frame_Psi_Left", _wt.frame_left.psi_install),
        XML_Node("Frame_U_Left", _wt.frame_left.u_value),
        XML_Node("Glazing_Psi_Left", _wt.frame_left.psi_glazing),
        # --
        XML_Node("Frame_Width_Right", _wt.frame_right.width),
        XML_Node("Frame_Psi_Right", _wt.frame_right.psi_install),
        XML_Node("Frame_U_Right", _wt.frame_right.u_value),
        XML_Node("Glazing_Psi_Right", _wt.frame_right.psi_glazing),
        # --
        XML_Node("Frame_Width_Top", _wt.frame_top.width),
        XML_Node("Frame_Psi_Top", _wt.frame_top.psi_install),
        XML_Node("Frame_U_Top", _wt.frame_top.u_value),
        XML_Node("Glazing_Psi_Top", _wt.frame_top.psi_glazing),
        # --
        XML_Node("Frame_Width_Bottom", _wt.frame_bottom.width),
        XML_Node("Frame_Psi_Bottom", _wt.frame_bottom.psi_install),
        XML_Node("Frame_U_Bottom", _wt.frame_bottom.u_value),
        XML_Node("Glazing_Psi_Bottom", _wt.frame_bottom.psi_glazing),
    ]


# -- VENTILATION --------------------------------------------------------------


def _PhxRoomVentilation(_r: loads.PhxRoomVentilation) -> List[xml_writable]:
    return [
        XML_Node('Name', _r.display_name),
        XML_Node('Type', _r.wufi_type),
        XML_Node('IdentNrUtilizationPatternVent', _r.vent_pattern_id_num),
        XML_Node('IdentNrVentilationUnit', _r.vent_unit_id_num),
        XML_Node('Quantity', _r.quantity),
        XML_Node('AreaRoom', _r.weighted_floor_area, "unit", "m²"),
        XML_Node('ClearRoomHeight', _r.clear_height, "unit", "m"),
        XML_Node('DesignVolumeFlowRateSupply',
                 round(_r.flow_rates.flow_supply, TOL_LEV1), "unit", "m³/h"),
        XML_Node('DesignVolumeFlowRateExhaust',
                 round(_r.flow_rates.flow_extract, TOL_LEV1), "unit", "m³/h"),
        # XML_Node('SupplyFlowRateUserDef', 'Test', "unit", "m³/h"),
        # XML_Node('ExhaustFlowRateUserDef', 'Test', "unit", "m³/h"),
        # XML_Node('DesignFlowInterzonalUserDef', 'Test', "unit", "m³/h"),
    ]


def _UtilizationPatternVent(_util_pat: schedules.UtilizationPatternVent) -> List[xml_writable]:
    op_periods = _util_pat.operating_periods
    return [
        XML_Node("Name", _util_pat.name),
        XML_Node("IdentNr", _util_pat.id_num),
        XML_Node("OperatingDays", _util_pat.operating_days),
        XML_Node("OperatingWeeks", _util_pat.operating_weeks),
        XML_Node("Maximum_DOS", round(
            op_periods.high.period_operating_hours, TOL_LEV1)),
        XML_Node("Maximum_PDF", round(
            op_periods.high.period_operation_speed, TOL_LEV1)),
        XML_Node("Standard_DOS", round(
            op_periods.standard.period_operating_hours, TOL_LEV1)),
        XML_Node("Standard_PDF", round(
            op_periods.standard.period_operation_speed, TOL_LEV1)),
        XML_Node("Basic_DOS", round(
            op_periods.basic.period_operating_hours, TOL_LEV1)),
        XML_Node("Basic_PDF", round(
            op_periods.basic.period_operation_speed, TOL_LEV1)),
        XML_Node("Minimum_DOS", round(
            op_periods.minimum.period_operating_hours, TOL_LEV1)),
        XML_Node("Minimum_PDF", round(
            op_periods.minimum.period_operation_speed, TOL_LEV1)),
    ]


# -- MECHANICAL DEVICES -------------------------------------------------------


def _PhxDeviceVentilator(_d: hvac.PhxDeviceVentilator) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Node("UseOptionalClimate", False),
        XML_Node("IdentNr_OptionalClimate", -1),
        XML_Node("HeatRecovery", _d.params.sensible_heat_recovery),
        XML_Node("MoistureRecovery ", _d.params.latent_heat_recovery),
        XML_Object('PH_Parameters', _d.params,
                   _schema_name='_DeviceVentilatorPhParams')
    ]


def _DeviceVentilatorPhParams(_p: hvac.PhxDeviceVentilatorParams) -> List[xml_writable]:
    return [
        XML_Node("Quantity", _p.quantity),
        XML_Node("HumidityRecoveryEfficiency", _p.latent_heat_recovery),
        XML_Node("ElectricEfficiency", _p.electric_efficiency),
        XML_Node("DefrostRequired", _p.frost_protection_reqd),
        XML_Node("FrostProtection", _p.frost_protection_reqd),
        XML_Node("TemperatureBelowDefrostUsed",
                 _p.temperature_below_defrost_used),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        # XML_Node("SubsoilHeatExchangeEfficiency", _p.),
        # XML_Node("VolumeFlowRateFrom", "unit","m³/h", _p.),
        # XML_Node("VolumeFlowRateTo", "unit","m³/h", _p.),
        # XML_Node("NoSummerBypass", _p.),
        # XML_Node("Maximum_VOS", _p.),
        # XML_Node("Maximum_PP", _p.),
        # XML_Node("Standard_VOS", _p.),
        # XML_Node("Standard_PP", _p.),
        # XML_Node("Basic_VOS", _p.),
        # XML_Node("Basic_PP", _p.),
        # XML_Node("Minimum_VOS", _p.),
        # XML_Node("Minimum_PP", _p.),
        # XML_Node("AuxiliaryEnergy", _p.),
        # XML_Node("AuxiliaryEnergyDHW", _p.),
    ]


# -- Elec Heating--------------------------------------------------------------


def _PhxDeviceHeaterElec(_d: hvac.PhxHeaterElectric) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Object('PH_Parameters', _d.params,
                   _schema_name='_DeviceHeaterElecPhParams'),
        XML_Object('DHW_Parameters', _d,
                   _schema_name='_DeviceHeaterElecDeviceParams'),
        XML_Object('Heating_Parameters', _d,
                   _schema_name='_DeviceHeaterElecDeviceParams'),
    ]


def _DeviceHeaterElecPhParams(_p: hvac.PhxMechanicalDeviceParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
    ]


def _DeviceHeaterElecDeviceParams(_d: hvac.PhxHeaterElectric) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.percent_coverage),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- Boilers ------------------------------------------------------------------


def _PhxDeviceHeaterBoiler(_d: hvac.PhxHeaterBoiler) -> List[xml_writable]:
    ph_params = {
        'GAS': '_DeviceHeaterBoilerFossilPhParams',
        'OIL': '_DeviceHeaterBoilerFossilPhParams',
        'WOOD_LOG': '_DeviceHeaterBoilerWoodPhParams',
        'WOOD_PELLET': '_DeviceHeaterBoilerWoodPhParams',
    }
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Object('PH_Parameters', _d.params,
                   _schema_name=ph_params[_d.params.fuel.name]),
        XML_Object('DHW_Parameters', _d,
                   _schema_name='_DeviceHeaterBoilerDeviceParams'),
        XML_Object('Heating_Parameters', _d,
                   _schema_name='_DeviceHeaterBoilerDeviceParams'),
    ]


def _DeviceHeaterBoilerWoodPhParams(_p: hvac.PhxHeaterBoilerWoodParams) -> List[xml_writable]:
    return [
        XML_Node("EnergySourceBoilerType", _p.fuel.value),
        XML_Node("MaximalBoilerPower", _p.rated_capacity),
        XML_Node("SolarFractionBoilerSpaceHeating", _p.solar_fraction),
        XML_Node("EfficiencyHeatGeneratorBasicCycle", _p.effic_in_basic_cycle),
        XML_Node("EfficiencyHeatGeneratorConstantOperation",
                 _p.effic_in_const_operation),
        XML_Node("AverageFractionHeatOutputReleasedHeatingCircuit",
                 _p.avg_frac_heat_output),
        XML_Node("TemperatureDifferencePowerOnPowerOff", _p.temp_diff_on_off),
        XML_Node("UsefulHeatOutputBasicCycl", _p.rated_capacity),
        XML_Node("AveragePowerOutputHeatGenerator", _p.rated_capacity),
        XML_Node("DemandBasicCycle", _p.demand_basic_cycle),
        XML_Node("PowerConsumptionStationarRun", _p.power_standard_run),
        XML_Node("NoTransportPellets", _p.no_transport_pellets),
        XML_Node("OnlyControl", _p.only_control),
        XML_Node("AreaMechanicalRoom", _p.area_mech_room),
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
    ]


def _DeviceHeaterBoilerFossilPhParams(_p: hvac.PhxHeaterBoilerFossilParams) -> List[xml_writable]:
    return [
        XML_Node("EnergySourceBoilerType", _p.fuel.value),
        XML_Node("CondensingBoiler", _p.condensing),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("MaximalBoilerPower", _p.rated_capacity),
        XML_Node("BoilerEfficiency30", _p.effic_at_30_percent_load),
        XML_Node("BoilerEfficiencyNominalOutput", _p.effic_at_nominal_load),
        XML_Node("AverageReturnTemperatureMeasured30Load",
                 _p.avg_rtrn_temp_at_30_percent_load),
        XML_Node("AverageBoilerTemperatureDesign70_55",
                 _p.avg_temp_at_70C_55C),
        XML_Node("AverageBoilerTemperatureDesign55_45",
                 _p.avg_temp_at_55C_45C),
        XML_Node("AverageBoilerTemperatureDesign35_28",
                 _p.avg_temp_at_32C_28C),
        XML_Node("StandbyHeatLossBoiler70", _p.standby_loss_at_70C),
        XML_Node("SolarFractionBoilerSpaceHeating", _p.aux_energy),
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
    ]


def _DeviceHeaterBoilerDeviceParams(_d: hvac.PhxHeaterBoiler) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.percent_coverage),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- District Heat ------------------------------------------------------------


def _PhxDeviceHeaterDistrict(_d: hvac.PhxHeaterDistrictHeat) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Object('DHW_Parameters', _d,
                   _schema_name='_DeviceHeaterDistrictDeviceParams'),
        XML_Object('Heating_Parameters', _d,
                   _schema_name='_DeviceHeaterDistrictDeviceParams'),
    ]


def _DeviceHeaterDistrictDeviceParams(_d: hvac.PhxHeaterDistrictHeat) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.percent_coverage),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


# -- Heat Pumps ---------------------------------------------------------------


def _PhxDeviceHeaterHeatPump(_d: hvac.PhxHeaterHeatPump) -> List[xml_writable]:
    param_schemas = {
        hvac.PhxHeaterHeatPumpAnnualParams.hp_type.value: '_DeviceHeaterHeatPumpPhParamsAnnual',
        hvac.PhxHeaterHeatPumpMonthlyParams.hp_type.value: '_DeviceHeaterHeatPumpPhParamsMonthly',
        hvac.PhxHeaterHeatPumpHotWaterParams.hp_type.value: '_DeviceHeaterHeatPumpPhParamsHotWater',
        hvac.PhxHeaterHeatPumpCombinedParams.hp_type.value: '_DeviceHeaterHeatPumpPhParamsCombined',
    }
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Object('PH_Parameters', _d.params,
                   _schema_name=param_schemas[_d.params.hp_type.value]),
        XML_Object('DHW_Parameters', _d,
                   _schema_name='_DeviceHeaterHeatPumpDeviceParams'),
        XML_Object('Heating_Parameters', _d,
                   _schema_name='_DeviceHeaterHeatPumpDeviceParams'),
    ]


def _DeviceHeaterHeatPumpPhParamsAnnual(_p: hvac.PhxHeaterHeatPumpAnnualParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("AnnualCOP", _p.annual_COP),
        XML_Node("TotalSystemPerformanceRatioHeatGenerator",
                 _p.total_system_perf_ratio),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsMonthly(_p: hvac.PhxHeaterHeatPumpMonthlyParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("RatedCOP1", _p.COP_1),
        XML_Node("RatedCOP2", _p.COP_2),
        XML_Node("AmbientTemperature1", _p.ambient_temp_1),
        XML_Node("AmbientTemperature2", _p.ambient_temp_2),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsHotWater(_p: hvac.PhxHeaterHeatPumpHotWaterParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("AnnualCOP", _p.annual_COP),
        XML_Node("TotalSystemPerformanceRatioHeatGenerator",
                 _p.annual_system_perf_ratio),
        XML_Node("HPWH_EF", _p.annual_energy_factor),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpPhParamsCombined(_p: hvac.PhxHeaterHeatPumpCombinedParams) -> List[xml_writable]:
    return [
        XML_Node("AuxiliaryEnergy", _p.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _p.aux_energy_dhw),
        XML_Node("InConditionedSpace", _p.in_conditioned_space),
        XML_Node("HPType", _p.hp_type.value),
    ]


def _DeviceHeaterHeatPumpDeviceParams(_d: hvac.PhxHeaterHeatPump) -> List[xml_writable]:
    return [
        XML_Node("CoverageWithinSystem", _d.percent_coverage),
        XML_Node("Unit", _d.unit),
        XML_Node("Selection", 1),
    ]


def _PhxDeviceWaterStorage(_d: hvac.PhxHotWaterTank) -> List[xml_writable]:
    return [
        XML_Node("Name", _d.display_name),
        XML_Node("IdentNr", _d.id_num),
        XML_Node("SystemType", _d.system_type.value),
        XML_Node("TypeDevice", _d.device_type.value),
        XML_Node("UsedFor_Heating", _d.usage_profile.space_heating),
        XML_Node("UsedFor_DHW", _d.usage_profile.dhw_heating),
        XML_Node("UsedFor_Cooling", _d.usage_profile.cooling),
        XML_Node("UsedFor_Ventilation", _d.usage_profile.ventilation),
        XML_Node("UsedFor_Humidification", _d.usage_profile.humidification),
        XML_Node("UsedFor_Dehumidification",
                 _d.usage_profile.dehumidification),
        XML_Object('PH_Parameters', _d,
                   _schema_name='_DeviceWaterStoragePhParams')
    ]


def _DeviceWaterStoragePhParams(_t: hvac.PhxHotWaterTank) -> List[xml_writable]:
    return [
        XML_Node("SolarThermalStorageCapacity", _t.params.storage_capacity),
        XML_Node("StorageLossesStandby", _t.params.standby_losses),
        XML_Node("TotalSolarThermalStorageLosses", _t.params.standby_losses),
        XML_Node("InputOption", _t.params.input_option.value),
        XML_Node("AverageHeatReleaseStorage", _t.params.storage_loss_rate),
        XML_Node("TankRoomTemp ", _t.params.room_temp),
        XML_Node("TypicalStorageWaterTemperature", _t.params.water_temp),
        XML_Node("QauntityWS", _t.quantity),
        XML_Node("AuxiliaryEnergy", _t.params.aux_energy),
        XML_Node("AuxiliaryEnergyDHW", _t.params.aux_energy_dhw),
        XML_Node("InConditionedSpace", _t.params.in_conditioned_space),
    ]


# -- MECHANICAL SYSTEMS / DISTRIBUTION ----------------------------------------


class DistributionDHW:
    def __init__(self):
        raise NotImplementedError


def _DistributionDHW(_d):
    raise NotImplementedError
    # return [
    #     XML_Node("LengthCirculationPipes_WR", _d.),
    #     XML_Node("LengthCirculationPipes_CR1", _d.),
    #     XML_Node("LengthCirculationPipes_CR2", _d.),
    #     XML_Node("HeatLossCoefficient_WR", _d.),
    #     XML_Node("HeatLossCoefficient_CR1", _d.),
    #     XML_Node("HeatLossCoefficient_CR2", _d.),
    #     XML_Node("TemperatureRoom_WR", _d.),
    #     XML_Node("TemperatureRoom_CR1", _d.),
    #     XML_Node("TemperatureRoom_CR2", _d.),
    #     XML_Node("DesignFlowTemperature_WR", _d.),
    #     XML_Node("DesignFlowTemperature_CR1", _d.),
    #     XML_Node("DesignFlowTemperature_CR2", _d.),
    #     XML_Node("DailyRunningHoursCirculation_WR", _d.),
    #     XML_Node("DailyRunningHoursCirculation_CR1", _d.),
    #     XML_Node("DailyRunningHoursCirculation_CR2", _d.),
    #     XML_Node("LengthIndividualPipes_WR", _d.),
    #     XML_Node("LengthIndividualPipes_CR1", _d.),
    #     XML_Node("LengthIndividualPipes_CR2", _d.),
    #     XML_Node("ExteriorPipeDiameter_WR", _d.),
    #     XML_Node("ExteriorPipeDiameter_CR1", _d.),
    #     XML_Node("ExteriorPipeDiameter_CR2", _d.),
    #     XML_Node("HeatReleaseStorage_WR", _d.),
    #     XML_Node("HeatReleaseStorage_CR1", _d.),
    #     XML_Node("HeatReleaseStorage_CR2", _d.),
    #     XML_Node("CalculationMethodIndividualPipes", _d.),
    #     XML_Node("PipeMaterialSimplifiedMethod", _d.),
    #     XML_Node("PipeDiameterSimplifiedMethod", _d.),
    #     XML_Node("HotWaterFixtureEffectiveness", _d.),
    #     XML_Node("DemandRecirculation", _d.),
    #     XML_Node("SelectionhotWaterFixtureEff", _d.),
    #     XML_Node("NumberOfBathrooms", _d.),
    #     XML_Node("AllPipesAreInsulated", _d.),
    #     XML_Node("SelectionUnitsOrFloors", _d.),
    # ]


class DistributionHeating:
    def __init__(self):
        raise NotImplementedError


def _DistributionHeating(_d):
    raise NotImplementedError
    # return [
    #     XML_Node("LengthPipes_WR", _d.),
    #     XML_Node("LengthPipes_CR1", _d.),
    #     XML_Node("LengthPipes_CR2", _d.),
    #     XML_Node("HeatLossCoefficient_WR", _d.),
    #     XML_Node("HeatLossCoefficient_CR1", _d.),
    #     XML_Node("HeatLossCoefficient_CR2", _d.),
    #     XML_Node("TemperatureRoom_WR", _d.),
    #     XML_Node("TemperatureRoom_CR1", _d.),
    #     XML_Node("TemperatureRoom_CR2", _d.),
    #     XML_Node("DesignFlowTemperature_WR", _d.),
    #     XML_Node("DesignFlowTemperature_CR1", _d.),
    #     XML_Node("DesignFlowTemperature_CR2", _d.),
    #     XML_Node("DesignSystemHeatingLoad_WR", _d.),
    #     XML_Node("DesignSystemHeatingLoad_CR1", _d.),
    #     XML_Node("DesignSystemHeatingLoad_CR2", _d.),
    #     XML_Node("FlowTControl_WR", _d.),
    #     XML_Node("FlowTControl_CR1", _d.),
    #     XML_Node("FlowTControl_CR2", _d.),
    # ]


def _Duct(_in):
    return [
        # XML_Node("Name", _d.),
        # XML_Node("IdentNr", _d.),
        # XML_Node("DuctDiameter", _d.),
        # XML_Node("DuctLength", _d.),
        # XML_Node("InsulationThickness", _d.),
        # XML_Node("ThermalConductivity", _d.),
        # XML_Node("Quantity", _d.),
        # XML_Node("DuctType", _d.),
        # XML_Node("DuctShape", _d.),
        # XML_Node("IsReflective", _d.),
        # XML_Node("AssignedVentUnits", _d.),
        # XML_Node("IdentNrVentUnit", _d.),
        # XML_Node("/AssignedVentUnits", _d.),
    ]


"""DEV NOTE: I don't want to have the Cooling distribution as part of the PHX model.
It is stupid that things like COP are stored in there. So use a temp class for now."""


class TempDistributionCooling:
    """Temporary wrapper class for WUFI format Cooling Distribution data"""

    def __init__(self, _devices: List[hvac.AnyPhxCooling]):
        # -- have to sort and combine the systems together       
        self.ventilation_device = self.sum_devices(_devices, hvac.CoolingType.VENTILATION)
        self.recirculation_device = self.sum_devices(_devices, hvac.CoolingType.RECIRCULATION)
        self.dehumidification_device = self.sum_devices(_devices, hvac.CoolingType.DEHUMIDIFICATION)
        self.panel_device = self.sum_devices(_devices, hvac.CoolingType.PANEL)

    def sum_devices(self, _cooling_devices: List[hvac.AnyPhxCooling], _device_type: hvac.CoolingType) -> Optional[hvac.AnyPhxCooling]:
        """ Returns a single HVAC Cooling device, made from a list of input devices, or None if no devices found.
        
        Arguments:
        ----------
            * _cooling_devices: 
            * _device_type: 
        
        Returns:
        --------
            * Optional[hvac.AnyPhxCooling]: 
        """
        devices = [d for d in _cooling_devices if d.cooling_type == _device_type]
        if len(devices) == 0:
            return None
        else:
            return sum(devices)


def _DistributionCooling(_clg_distr: TempDistributionCooling) -> List[xml_writable]:
    base = []

    if _clg_distr.ventilation_device:
        vent_params: hvac.PhxCoolingVentilationParams = _clg_distr.ventilation_device.params
        base += [
            XML_Node("CoolingViaVentilationAir", True),
            XML_Node("SupplyAirCoolingOnOff", vent_params.single_speed),
            XML_Node("MaxSupplyAirCoolingPower", vent_params.capacity),
            XML_Node("MinTemperatureCoolingCoilSupplyAir",
                     vent_params.min_coil_temp),
            XML_Node("SupplyAirCoolinCOP", vent_params.annual_COP),
        ]
    if _clg_distr.recirculation_device:
        recirc_params: hvac.PhxCoolingRecirculationParams = _clg_distr.recirculation_device.params
        base += [
            XML_Node("CoolingViaRecirculation", True),
            XML_Node("RecirculatingAirOnOff", recirc_params.single_speed),
            XML_Node("MaxRecirculationAirCoolingPower",
                     recirc_params.capacity),
            XML_Node("MinTempCoolingCoilRecirculatingAir",
                     recirc_params.min_coil_temp),
            XML_Node("RecirculationCoolingCOP", recirc_params.annual_COP),
            XML_Node("RecirculationAirVolume", recirc_params.flow_rate_m3_hr),
            XML_Node("ControlledRecirculationVolumeFlow",
                     recirc_params.flow_rate_variable),
        ]
    if _clg_distr.dehumidification_device:
        dehumid_params: hvac.PhxCoolingDehumidificationParams = _clg_distr.dehumidification_device.params
        base += [
            XML_Node("Dehumidification", True),
            XML_Node("UsefullDehumidificationHeatLoss",
                     dehumid_params.useful_heat_loss),
            XML_Node("DehumdificationCOP", dehumid_params.annual_COP),
            XML_Node("SEER", None),
            XML_Node("EER", None),
            XML_Node("DehumidificationElEnergy", None),
        ]
    if _clg_distr.panel_device:
        panel_params: hvac.PhxCoolingPanelParams = _clg_distr.panel_device.params
        base += [
            XML_Node("PanelCooling", True),
            XML_Node("DehumdificationCOP", panel_params.annual_COP),
        ]
    
    return base


def _PHDistribution(_c: hvac.PhxMechanicalSystemCollection):
    return [
        # XML_Object('DistributionDHW', DistributionDHW()),
        # XML_Object('DistributionHeating', DistributionHeating()),
        XML_Object('DistributionCooling', 
            TempDistributionCooling(_c.cooling_devices), _schema_name='_DistributionCooling'
        ),
        # XML_List('DistributionVentilation', [XML_Object('Duct', None, 'index', i)
        #                                      for i, d in enumerate([])]),
        XML_Node("UseDefaultValues", True),
        XML_Node("DeviceInConditionedSpace", True),
    ]


# -- MECHANICAL SYSTEMS / COLLECTIONS -----------------------------------------


def _PhxZoneCoverage(_zc: hvac.PhxZoneCoverage) -> List[xml_writable]:
    return [
        XML_Node("IdentNrZone", _zc.zone_num),
        XML_Node("CoverageHeating", _zc.zone_num),
        XML_Node("CoverageCooling", _zc.zone_num),
        XML_Node("CoverageVentilation", _zc.zone_num),
        XML_Node("CoverageHumidification", _zc.zone_num),
        XML_Node("CoverageDehumidification", _zc.zone_num),
    ]


def _PhxMechanicalDevices(_hvac_collection: hvac.PhxMechanicalSystemCollection) -> List[xml_writable]:
    devices = {
        hvac.DeviceType.VENTILATION: '_PhxDeviceVentilator',
        hvac.DeviceType.ELECTRIC: '_PhxDeviceHeaterElec',
        hvac.DeviceType.BOILER: '_PhxDeviceHeaterBoiler',
        hvac.DeviceType.DISTRICT_HEAT: '_PhxDeviceHeaterDistrict',
        hvac.DeviceType.HEAT_PUMP: '_PhxDeviceHeaterHeatPump',
        hvac.DeviceType.WATER_STORAGE: '_PhxDeviceWaterStorage',
    }

    return [
        XML_Node("Name", _hvac_collection.display_name),
        XML_Node("Type", _hvac_collection.sys_type_num,
                 'choice', _hvac_collection.sys_type_str),
        XML_Node("IdentNr", _hvac_collection.id_num),
        XML_List('ZonesCoverage', [XML_Object("ZoneCoverage", n, "index", i)
                 for i, n in enumerate([_hvac_collection.zone_coverage])]),
        XML_List('Devices', [XML_Object("Device", d, "index", i, _schema_name=devices[d.device_type])
                 for i, d in enumerate(_hvac_collection.devices)]),
        XML_Object('PHDistribution', _hvac_collection,
                   _schema_name='_PHDistribution'),
    ]


def _PhxMechanicalSystemCollection(_hvac: hvac.PhxMechanicalSystemCollection) -> List[xml_writable]:
    return [
        XML_List(
            "Systems",
            [
                XML_Object("System", n, "index", i, _schema_name='_PhxMechanicalDevices')
                for i, n in enumerate([_hvac])
            ]
        ),
    ]


# -- ELEC. EQUIPMENT DEVICES --------------------------------------------------


def _PhxDeviceDishwasher(_d: elec_equip.PhxDeviceDishwasher) -> List[xml_writable]:
    return [
        XML_Node('Type', 1),
        XML_Node('Connection', _d.water_connection),
        XML_Node('DishwasherCapacityPreselection', _d.capacity_type),
        XML_Node('DishwasherCapacityInPlace', _d.capacity),
    ]


def _PhxDeviceClothesWasher(_d: elec_equip.PhxDeviceClothesWasher) -> List[xml_writable]:
    return [
        XML_Node('Type', 2),
        XML_Node('Connection', _d.water_connection),
        XML_Node('UtilizationFactor', _d.utilization_factor),
        XML_Node('CapacityClothesWasher', _d.capacity),
        XML_Node('MEF_ModifiedEnergyFactor', _d.modified_energy_factor),
    ]


def _PhxDeviceClothesDryer(_d: elec_equip.PhxDeviceClothesDryer) -> List[xml_writable]:
    return [
        XML_Node('Type', 3),
        XML_Node('Dryer_Choice', _d.dryer_type),
        XML_Node('GasConsumption', _d.gas_consumption),
        XML_Node('EfficiencyFactorGas', _d.gas_efficiency_factor),
        XML_Node('FieldUtilizationFactorPreselection',
                 _d.field_utilization_factor_type),
        XML_Node('FieldUtilizationFactor', _d.field_utilization_factor),
    ]


def _PhxDeviceRefrigerator(_d: elec_equip.PhxDeviceRefrigerator) -> List[xml_writable]:
    return [
        XML_Node('Type', 4),
    ]


def _PhxDeviceFreezer(_d: elec_equip.PhxDeviceFreezer) -> List[xml_writable]:
    return [
        XML_Node('Type', 5),
    ]


def _PhxDeviceFridgeFreezer(_d: elec_equip.PhxDeviceFridgeFreezer) -> List[xml_writable]:
    return [
        XML_Node('Type', 6),
    ]


def _PhxDeviceCooktop(_d: elec_equip.PhxDeviceCooktop) -> List[xml_writable]:
    return [
        XML_Node('Type', 7),
        XML_Node('CookingWith', _d.cooktop_type),
    ]


def _PhxDeviceMEL(_d: elec_equip.PhxDeviceMEL) -> List[xml_writable]:
    return [
        XML_Node('Type', 13),
    ]


def _PhxDeviceLightingInterior(_d: elec_equip.PhxDeviceLightingInterior) -> List[xml_writable]:
    return [
        XML_Node('Type', 14),
        XML_Node('FractionHightEfficiency', _d.frac_high_efficiency),
    ]


def _PhxDeviceLightingExterior(_d: elec_equip.PhxDeviceLightingExterior) -> List[xml_writable]:
    return [
        XML_Node('Type', 15),
        XML_Node('FractionHightEfficiency', _d.frac_high_efficiency),
    ]


def _PhxDeviceLightingGarage(_d: elec_equip.PhxDeviceLightingGarage) -> List[xml_writable]:
    return [
        XML_Node('Type', 16),
        XML_Node('FractionHightEfficiency', _d.frac_high_efficiency),
    ]


def _PhxDeviceCustomElec(_d: elec_equip.PhxDeviceCustomElec) -> List[xml_writable]:
    return [
        XML_Node('Type', 11),
    ]


def _PhxDeviceCustomLighting(_d: elec_equip.PhxDeviceCustomLighting) -> List[xml_writable]:
    return [
        XML_Node('Type', 17),
    ]


def _PhxDeviceCustomMEL(_d: elec_equip.PhxDeviceCustomMEL) -> List[xml_writable]:
    return [
        XML_Node('Type', 18),
    ]


def _PhxElectricalDevice(_d: elec_equip.PhxElectricalDevice) -> List[xml_writable]:
    common_attributes = [
        XML_Node('Comment', _d.comment),
        XML_Node('ReferenceQuantity', _d.reference_quantity),
        XML_Node('Quantity', _d.quantity),
        XML_Node('InConditionedSpace', _d.in_conditioned_space),
        XML_Node('ReferenceEnergyDemandNorm', _d.reference_energy_norm),
        XML_Node('EnergyDemandNorm', _d.energy_demand),
        XML_Node('EnergyDemandNormUse', _d.energy_demand_per_use),
        XML_Node('CEF_CombinedEnergyFactor', _d.combined_energy_factor),
    ]

    device_schema = getattr(sys.modules[__name__], f'_{_d.__class__.__name__}')
    appliance_specific_attributes = device_schema(_d)
    return common_attributes + appliance_specific_attributes



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create PHX-Service Hot Water objects from Honeybee-Energy-PH SHW"""

from PHX.model import hvac
from honeybee_energy_ph.hvac import hot_water
from PHX.model.enums.hvac import PhxHotWaterTankType
from PHX.model.hvac import piping

# -- Storage -- 

def build_phx_hw_tank(_hbph_tank: hot_water.PhSHWTank) -> hvac.PhxHotWaterTank:
    """Returns a new PHX Hot-Water Tank based on the Honeybee-PH Hot Water Tank input.

    Arguments:
    ----------
        * _hbph_heater (hot_water.PhSHWTank): The Honeybee-PH Hot-Water tank
            to use as the source.

    Returns:
    --------
        * mech_equip.PhxHotWaterTank: The new PHX-Hot-Water-Tank.
    """

    phx_tank = hvac.PhxHotWaterTank()

    phx_tank.display_name = _hbph_tank.display_name
    phx_tank.quantity = _hbph_tank.quantity

    phx_tank.params.tank_type = PhxHotWaterTankType.from_hbph_type(_hbph_tank.tank_type)
    phx_tank.params.in_conditioned_space = _hbph_tank.in_conditioned_space
    
    phx_tank.params.solar_connection = _hbph_tank.solar_connection
    phx_tank.params.solar_losses = _hbph_tank.solar_losses    

    phx_tank.params.storage_capacity = _hbph_tank.storage_capacity
    phx_tank.params.storage_loss_rate = _hbph_tank.storage_loss_rate

    phx_tank.params.standby_losses = _hbph_tank.standby_losses
    phx_tank.params.standby_fraction = _hbph_tank.standby_fraction

    phx_tank.params.room_temp = _hbph_tank.room_temp
    phx_tank.params.water_temp = _hbph_tank.water_temp

    return phx_tank


def build_phx_hw_storage(_hbph_tank: hot_water.PhSHWTank) -> hvac.PhxHotWaterTank:
    """

    Arguments:
    ----------
        * _hbph_heater (hot_water.PhSHWTank): The Honeybee-PH Hot-Water tank
            to use as the source.

    Returns:
    --------
        * (mech.PhxHotWaterTank): The new Water Storage SubSystem.
    """

    phx_storage_tank = build_phx_hw_tank(_hbph_tank)

    # TODO: Distribution...

    return phx_storage_tank

# -- Heaters ----

def build_phx_hw_heater(_hbph_heater: hot_water.PhHotWaterHeater) -> hvac.PhxHeatingDevice:
    """Returns a new PHX Hot-Water Heater based on the Honeybee-PH Hot Water Heater input.

    Arguments:
    ----------
        * _hbph_heater (hot_water.PhHotWaterHeater): The Honeybee-PH Hot-Water heater
            to use as the source for the PHX Heater.

    Returns:
    --------
        * (hvac.PhxHeatingDevice): The new PHX-Hot-Water-Heater.
    """

    # -- Get the right constructor based on the type of heater
    heaters = {
        'PhSHWHeaterElectric': hvac.PhxHeaterElectric,
        'PhSHWHeaterBoiler': hvac.PhxHeaterBoilerFossil,
        'PhSHWHeaterBoilerWood': hvac.PhxHeaterBoilerWood,
        'PhSHWHeaterDistrict': hvac.PhxHeaterDistrictHeat,
        'PhSHWHeaterHeatPump': hvac.PhxHeaterHeatPumpHotWater,
    }

    # -- Build the basic heater and set basic data
    heater_class = heaters[_hbph_heater.__class__.__name__]
    phx_hw_heater = heater_class()
    phx_hw_heater.display_name = _hbph_heater.display_name

    # -- Pull out all the detailed data which varies depending on the 'type'
    for attr_name in vars(phx_hw_heater).keys():
        try:
            if attr_name.startswith('_'):
                attr_name = attr_name[1:]
            setattr(phx_hw_heater, attr_name, getattr(_hbph_heater, attr_name))
        except AttributeError:
            pass

    for attr_name in vars(phx_hw_heater.params).keys():
        try:
            if attr_name.startswith('_'):
                attr_name = attr_name[1:]
            setattr(phx_hw_heater.params, attr_name, getattr(_hbph_heater, attr_name))
        except AttributeError:
            pass

    phx_hw_heater.usage_profile.dhw_heating = True

    return phx_hw_heater

# -- Piping ----

def build_phx_piping(_hbph_pipe: hot_water.PhPipeElement) -> hvac.PhxPipeElement:
    """Create a new PHX Water Pipe based on an input Honeybee-PH Water Pipe.
    
    Arguments:
    ----------
        * _hbph_pipe: (hot_water.PhPipeElement): The Honeybee-PH Pipe to use as 
            the source.
    
    Returns:
    --------
        * (hvac.PhxPipeElement): The new PhxPipeElement created.
    """

    phx_pipe = piping.PhxPipeElement(
        _hbph_pipe.identifier,
        _hbph_pipe.display_name
    )
    for segment in _hbph_pipe.segments:
        phx_pipe.add_segment(
            piping.PhxPipeSegment(
                segment.identifier,
                segment.display_name,
                segment.geometry,
                segment.diameter,
                segment.insulation_thickness,
                segment.insulation_conductivity,
                segment.insulation_reflective,
                segment.insulation_quality,
                segment.daily_period,
            )
        )

    return phx_pipe



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP DHW Tank"""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model.hvac import piping

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class RecircPipingInput:
    """Model class for a single DHW Recirculation Pipe Element."""

    __slots__ = ('shape', 'phx_pipe', 'pipe_group_num')
    shape: shape_model.Dhw
    phx_pipe: List[piping.PhxPipeSegment]
    pipe_group_num: int
    
    @property
    def input_column(self) -> str:
        """Return the right input column based on the pipe-group-number."""
        return xl_data.col_offset(self.shape.recirc_piping.input_col_start, self.pipe_group_num)
        
    def _create_range(self, _field_name: str, _row_num: int, ) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f'{self.input_column}{_row_num + getattr(self.shape.recirc_piping.input_rows_offset, _field_name).row}'
    
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.recirc_piping.input_rows_offset, _field_name).unit

    def _bool_as_x(self, _value: bool):
        if _value == True:
            return 'x'
        else:
            return ''

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemDHW = partial(xl_data.XlItem, _sheet_name)
        total_length = sum(s.length for s in self.phx_pipe)
        
        return [
            XLItemDHW(
                    create_range('total_length'),
                    total_length,
                    "M", 
                    self._get_target_unit('total_length')
            ),
            XLItemDHW(
                    create_range('diameter'),
                    sum(s.diameter * s.length for s in self.phx_pipe) / total_length,
                    "M", 
                    self._get_target_unit('diameter')
            ),
            XLItemDHW(
                    create_range('insul_thickness'),
                    sum(s.insulation_thickness * s.length for s in self.phx_pipe) / total_length,
                    "M", 
                    self._get_target_unit('insul_thickness')
            ),
            XLItemDHW(
                    create_range('insul_reflective'),
                    self._bool_as_x(any([s.insulation_reflective for s in self.phx_pipe])),
            ),
            XLItemDHW(
                    create_range('insul_conductivity'),
                    sum(s.insulation_thickness * s.length for s in self.phx_pipe) / total_length,
                    "W/MK", 
                    self._get_target_unit('insul_conductivity')
            ),
        ]


@dataclass
class BranchPipingInput:
    """Model class for a single DHW Branch Pipe Element."""

    __slots__ = ('shape', 'phx_pipe', 'pipe_group_num', 'num_tap_points')
    shape: shape_model.Dhw
    phx_pipe: List[piping.PhxPipeSegment]
    pipe_group_num: int
    num_tap_points: int
    
    @property
    def input_column(self) -> str:
        """Return the right input column based on the pipe-group-number."""
        return xl_data.col_offset(self.shape.branch_piping.input_col_start, self.pipe_group_num)

    def create_range(self, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f'{self.input_column}{_row_num}'
        
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.branch_piping.input_rows_offset, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of Branch Piping Xl-Write items."""
        
        XLItemDHW = partial(xl_data.XlItem, _sheet_name)
        return [
            # -- Branch Piping
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.diameter.row), # type: ignore
                sum(s.diameter * s.length for s in self.phx_pipe) / sum(s.length for s in self.phx_pipe),
                    "M", 
                    self._get_target_unit('diameter')
            ),
            XLItemDHW(
                    self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.total_length.row), # type: ignore
                    sum(s.length for s in self.phx_pipe),
                    "M", 
                    self._get_target_unit('total_length')
                ),
            # -- Add Tapping Points
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.num_taps),
                self.num_tap_points,
            ),
        ]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Electricity / Equipment row input."""

from dataclasses import dataclass
from typing import List, Tuple, Any

from PHX.model import elec_equip

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class ElectricityItem:
    """Model class for a single Electric-Equipment item entry row."""

    __slots__ = ('phx_equipment')
    phx_equipment: Any

    def create_xl_items(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        """Returns a list of xl_data.XlItem or raises and Error if equipment type is unrecognized."""
        xl_item_functions = {
            'PhxDeviceDishwasher': self._dishwasher,
            'PhxDeviceClothesWasher': self._clothes_washer,
            'PhxDeviceClothesDryer': self._clothes_dryer,
            'PhxDeviceRefrigerator': self._refrigerator,
            'PhxDeviceFreezer': self._freezer,
            'PhxDeviceFridgeFreezer': self._fridge_freezer,
            'PhxDeviceCooktop': self._cooktop,
            'PhxDeviceMEL': self._mel,
            'PhxDeviceLightingInterior': self._lighting_interior,
            'PhxDeviceLightingExterior': self._lighting_exterior,
            'PhxDeviceLightingGarage': self._lighting_garage,
            'PhxDeviceCustomElec': self._custom_elec,
            'PhxDeviceCustomLighting': self._custom_lighting,
            'PhxDeviceCustomMEL': self._custom_mel,
        }
        try:
            return xl_item_functions[self.phx_equipment.__class__.__name__](_shape)
        except KeyError:
            raise NotImplementedError(
                f"No matching XL-write function found for equipment type: '{self.phx_equipment.__class__.__name__}'"
            )    
    
    def _dishwasher(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceDishwasher = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f'{_shape.input_columns.used}{_shape.input_rows.dishwasher.data}',
                1
            ),
            (
                f'{_shape.input_columns.in_conditioned_space}{_shape.input_rows.dishwasher.data}',
                str(int(equip.in_conditioned_space))
            ),            
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.dishwasher.data}',
                equip.energy_demand_per_use
            ),
            (
                f'{_shape.input_columns.selection}{_shape.input_rows.dishwasher.selection}',
                _shape.input_rows.dishwasher.selection_options[str(equip.water_connection)]
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]
    
    def _clothes_washer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceClothesWasher = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f'{_shape.input_columns.used}{_shape.input_rows.clothes_washing.data}',
                1
            ),
            (
                f'{_shape.input_columns.in_conditioned_space}{_shape.input_rows.clothes_washing.data}',
                str(int(equip.in_conditioned_space))
            ),            
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_washing.data}',
                equip.energy_demand_per_use
            ),
            (
                f'{_shape.input_columns.selection}{_shape.input_rows.clothes_washing.selection}',
                _shape.input_rows.clothes_washing.selection_options[str(equip.water_connection)]
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]
    
    def _clothes_dryer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceClothesDryer = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
             (
                f'{_shape.input_columns.in_conditioned_space}{_shape.input_rows.clothes_drying.data}',
                str(int(equip.in_conditioned_space))
            ),
            (
                f'{_shape.input_columns.selection}{_shape.input_rows.clothes_drying.selection}',
                _shape.input_rows.clothes_drying.selection_options[str(equip.dryer_type)]
            ),
        ]

        # -- Add energy consumption, location depends on fuel type
        if equip.dryer_type == 6:
            # Gas dryer
            items.append(
                (                
                    f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_drying.selection}',
                    equip.gas_consumption
                )
            )
        else: 
            # non-Gas dryer
            items.append(
                (
                    f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_drying.data}',
                    equip.energy_demand_per_use
                ),
            )

        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _refrigerator(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceRefrigerator = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f'{_shape.input_columns.used}{_shape.input_rows.refrigerator.data}',
                1
            ),      
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.refrigerator.data}',
                equip.energy_demand_per_use
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _fridge_freezer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceFridgeFreezer = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f'{_shape.input_columns.used}{_shape.input_rows.fridge_freezer.data}',
                1
            ),       
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.fridge_freezer.data}',
                equip.energy_demand_per_use
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _freezer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceFreezer = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f'{_shape.input_columns.used}{_shape.input_rows.freezer.data}',
                1
            ),
            (
                f'{_shape.input_columns.in_conditioned_space}{_shape.input_rows.freezer.data}',
                str(int(equip.in_conditioned_space))
            ),            
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.freezer.data}',
                equip.energy_demand_per_use
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _cooktop(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceCooktop = self.phx_equipment      
        items: List[Tuple[str, xl_data.xl_writable]] = [      
            (
                f'{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.cooking.data}',
                equip.energy_demand_per_use
            ),
            (
                f'{_shape.input_columns.selection}{_shape.input_rows.cooking.selection}',
                _shape.input_rows.cooking.selection_options[str(equip.cooktop_type)]
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _mel(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []
        
    def _lighting_interior(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _lighting_exterior(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _lighting_garage(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_elec(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_lighting(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_mel(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []